# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import astropy.units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from astropy.table import Table
import numpy as np
from gammapy.data import Observation, ObservationsEventsSampler
from gammapy.data.pointing import FixedPointingInfo
from gammapy.datasets.tests.test_simulate import get_energy_dependent_temporal_model
from gammapy.irf import load_irf_dict_from_file
from gammapy.modeling.models import (
    ConstantSpectralModel,
    Models,
    GaussianSpatialModel,
    PointSpatialModel,
    PowerLawSpectralModel,
    LightCurveTemplateTemporalModel,
    SkyModel,
)
from gammapy.utils.testing import requires_data

LOCATION = EarthLocation(lon="-70d18m58.84s", lat="-24d41m0.34s", height="2000m")


@pytest.fixture()
def signal_model():
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
    )

    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")

    t_max = 1000 * u.s

    time = np.arange(t_max.value) * u.s
    tau = u.Quantity("2e2 s")
    norm = np.exp(-time / tau)

    table = Table()
    table["TIME"] = time
    table["NORM"] = norm / norm.max()
    t_ref = Time("2000-01-01")
    table.meta = dict(MJDREFI=t_ref.mjd, MJDREFF=0, TIMEUNIT="s", TIMESYS="utc")
    temporal_model = LightCurveTemplateTemporalModel.from_table(table)

    return SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model,
        temporal_model=temporal_model,
        name="test-source",
    )


@pytest.fixture()
def energy_dependent_temporal_sky_model():
    spatial_model = PointSpatialModel(lon_0="0 deg", lat_0="0 deg", frame="galactic")
    spectral_model = ConstantSpectralModel(const="1 cm-2 s-1 TeV-1")
    temporal_model = get_energy_dependent_temporal_model()
    model = SkyModel(
        spectral_model=spectral_model,
        spatial_model=spatial_model,
        temporal_model=temporal_model,
    )
    return model


@pytest.fixture(scope="session")
def observations():
    pointing = FixedPointingInfo(fixed_icrs=SkyCoord(0 * u.deg, 0 * u.deg))
    livetime = 0.5 * u.hr
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-caldb/Prod5-South-20deg-AverageAz-14MSTs37SSTs.180000s-v0.1.fits.gz"
    )
    observations = [
        Observation.create(
            obs_id=100 + k, pointing=pointing, livetime=livetime, irfs=irfs
        )
        for k in range(2)
    ]
    return observations


@pytest.fixture(scope="session")
def models_list():
    spectral_model_pwl = PowerLawSpectralModel(
        index=2, amplitude="1e-12 TeV-1 cm-2 s-1", reference="1 TeV"
    )
    spatial_model_point = PointSpatialModel(
        lon_0="0 deg", lat_0="0.0 deg", frame="galactic"
    )

    sky_model_pntpwl = SkyModel(
        spectral_model=spectral_model_pwl,
        spatial_model=spatial_model_point,
        name="point-pwl",
    )
    models = Models(sky_model_pntpwl)
    return models


@requires_data()
def test_observations_events_sampler(tmpdir, observations):
    sampler_kwargs = dict(random_state=0)
    dataset_kwargs = dict(
        spatial_bin_size_min=0.1 * u.deg,
        spatial_width_max=0.2 * u.deg,
        energy_bin_per_decade_max=2,
    )
    sampler = ObservationsEventsSampler(
        sampler_kwargs=sampler_kwargs,
        dataset_kwargs=dataset_kwargs,
        n_jobs=1,
        outdir=tmpdir,
        overwrite=True,
    )
    sampler.run(observations, models=None)


@requires_data()
def test_observations_events_sampler_time(
    tmpdir, observations, energy_dependent_temporal_sky_model
):
    models = Models(energy_dependent_temporal_sky_model)
    sampler_kwargs = dict(random_state=0)
    dataset_kwargs = dict(
        spatial_bin_size_min=0.1 * u.deg,
        spatial_width_max=0.2 * u.deg,
        energy_bin_per_decade_max=2,
    )
    sampler = ObservationsEventsSampler(
        sampler_kwargs=sampler_kwargs,
        dataset_kwargs=dataset_kwargs,
        n_jobs=1,
        outdir=tmpdir,
        overwrite=True,
    )
    sampler.run(observations, models=models)


@requires_data()
def test_observations_events_sampler_parallel(tmpdir, observations, models_list):
    sampler_kwargs = dict(random_state=0)
    dataset_kwargs = dict(
        spatial_bin_size_min=0.1 * u.deg,
        spatial_width_max=0.2 * u.deg,
        energy_bin_per_decade_max=2,
    )
    sampler = ObservationsEventsSampler(
        sampler_kwargs=sampler_kwargs,
        dataset_kwargs=dataset_kwargs,
        n_jobs=2,
        outdir=tmpdir,
        overwrite=True,
    )
    sampler.run(observations, models=models_list)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from pathlib import Path
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion
from pydantic import ValidationError
from gammapy.datasets import MapDataset, SpectrumDatasetOnOff
from gammapy.maps import WcsGeom, WcsNDMap
from gammapy.modeling.models import DatasetModels
from gammapy.utils.testing import requires_data
from gammapy.workflow import Workflow, WorkflowConfig

CONFIG_PATH = Path(__file__).resolve().parent / ".." / "config"
MODEL_FILE = CONFIG_PATH / "model.yaml"
MODEL_FILE_1D = CONFIG_PATH / "model-1d.yaml"


def get_example_config(which):
    """Example config: which can be 1d or 3d."""
    return WorkflowConfig.read(CONFIG_PATH / f"example-{which}.yaml")


def test_init():
    cfg = {"general": {"outdir": "test"}}
    workflow = Workflow(cfg)
    assert workflow.config.general.outdir == "test"
    with pytest.raises(TypeError):
        Workflow("spam")


def test_update_config():
    workflow = Workflow(WorkflowConfig())
    data = {"general": {"outdir": "test"}}
    config = WorkflowConfig(**data)
    workflow.update_config(config)
    assert workflow.config.general.outdir == "test"

    workflow = Workflow(WorkflowConfig())
    data = """
    general:
        outdir: test
    """
    workflow.update_config(data)
    assert workflow.config.general.outdir == "test"

    workflow = Workflow(WorkflowConfig())
    with pytest.raises(TypeError):
        workflow.update_config(0)


def test_get_observations_no_datastore():
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.datastore = "other"
    with pytest.raises(FileNotFoundError):
        workflow.get_observations()


@requires_data()
def test_get_observations_all():
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.datastore = "$GAMMAPY_DATA/cta-1dc/index/gps/"
    workflow.get_observations()
    assert len(workflow.observations) == 4


@requires_data()
def test_get_observations_obs_ids():
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.datastore = "$GAMMAPY_DATA/cta-1dc/index/gps/"
    workflow.config.observations.obs_ids = ["110380"]
    workflow.get_observations()
    assert len(workflow.observations) == 1


@requires_data()
def test_get_observations_obs_cone():
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.datastore = "$GAMMAPY_DATA/hess-dl3-dr1"
    workflow.config.observations.obs_cone = {
        "frame": "icrs",
        "lon": "83d",
        "lat": "22d",
        "radius": "5d",
    }
    workflow.get_observations()
    assert len(workflow.observations) == 4


@requires_data()
def test_get_observations_obs_file(tmp_path):
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.get_observations()
    filename = tmp_path / "obs_ids.txt"
    filename.write_text("20136\n47829\n")
    workflow.config.observations.obs_file = filename
    workflow.get_observations()
    assert len(workflow.observations) == 2


@requires_data()
def test_get_observations_obs_time(tmp_path):
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.obs_time = {
        "start": "2004-03-26",
        "stop": "2004-05-26",
    }
    workflow.get_observations()
    assert len(workflow.observations) == 40
    workflow.config.observations.obs_ids = [0]
    with pytest.raises(KeyError):
        workflow.get_observations()


@requires_data()
def test_get_observations_missing_irf():
    config = WorkflowConfig()
    workflow = Workflow(config)
    workflow.config.observations.datastore = "$GAMMAPY_DATA/joint-crab/dl3/magic/"
    workflow.config.observations.obs_ids = ["05029748"]
    workflow.config.observations.required_irf = ["aeff", "edisp"]
    workflow.get_observations()
    assert len(workflow.observations) == 1


@requires_data()
def test_set_models():
    config = get_example_config("3d")
    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    models_str = Path(MODEL_FILE).read_text()
    workflow.set_models(models=models_str)
    assert isinstance(workflow.models, DatasetModels)
    assert len(workflow.models) == 2
    assert workflow.models.names == ["source", "stacked-bkg"]
    with pytest.raises(TypeError):
        workflow.set_models(0)

    new_source = workflow.models["source"].copy(name="source2")
    workflow.set_models(models=[new_source], extend=False)
    assert len(workflow.models) == 2
    assert workflow.models.names == ["source2", "stacked-bkg"]


@requires_data()
def test_workflow_1d():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523, 23526]
        obs_time: {
            start: [J2004.92654346, J2004.92658453, J2004.92663655],
            stop: [J2004.92658453, J2004.92663655, J2004.92670773]
        }
    datasets:
        type: 1d
        background:
            method: reflected
        geom:
            axes:
                energy_true: {min: 0.01 TeV, max: 300 TeV, nbins: 109}
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.11 deg}
        safe_mask:
            methods: [aeff-default, edisp-bias]
            parameters: {bias_percent: 10.0}
        containment_correction: false
    flux_points:
        energy: {min: 1 TeV, max: 50 TeV, nbins: 4}
    light_curve:
        energy_edges: {min: 1 TeV, max: 50 TeV, nbins: 1}
        time_intervals: {
            start: [J2004.92654346, J2004.92658453, J2004.92663655],
            stop: [J2004.92658453, J2004.92663655, J2004.92670773]
        }
    """
    config = get_example_config("1d")
    workflow = Workflow(config)
    workflow.update_config(cfg)
    workflow.get_observations()
    workflow.get_datasets()
    workflow.read_models(MODEL_FILE_1D)
    workflow.run_fit()
    workflow.get_flux_points()
    workflow.get_light_curve()

    assert len(workflow.datasets) == 3
    table = workflow.flux_points.data.to_table(sed_type="dnde")

    assert len(table) == 4
    dnde = table["dnde"].quantity
    assert dnde.unit == "cm-2 s-1 TeV-1"

    assert_allclose(dnde[0].value, 8.116854e-12, rtol=1e-2)
    assert_allclose(dnde[2].value, 3.444475e-14, rtol=1e-2)

    axis = workflow.light_curve.geom.axes["time"]
    assert axis.nbin == 3
    assert_allclose(axis.time_min.mjd, [53343.92, 53343.935, 53343.954])

    flux = workflow.light_curve.flux.data[:, :, 0, 0]
    assert_allclose(flux, [[1.688954e-11], [2.347870e-11], [1.604152e-11]], rtol=1e-4)


@requires_data()
def test_geom_workflow_1d():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523]
    datasets:
        type: 1d
        background:
            method: reflected
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.11 deg}
        geom:
            axes:
                energy: {min: 0.1 TeV, max: 30 TeV, nbins: 20}
                energy_true: {min: 0.03 TeV, max: 100 TeV, nbins: 50}
        containment_correction: false
    flux_points:
        energy: {min: 1 TeV, max: 50 TeV, nbins: 4}
    """
    config = get_example_config("1d")
    workflow = Workflow(config)
    workflow.update_config(cfg)
    workflow.get_observations()
    workflow.get_datasets()

    assert len(workflow.datasets) == 1

    axis = workflow.datasets[0].exposure.geom.axes["energy_true"]
    assert axis.nbin == 50
    assert_allclose(axis.edges[0].to_value("TeV"), 0.03)
    assert_allclose(axis.edges[-1].to_value("TeV"), 100)


@requires_data()
def test_exclusion_region(tmp_path):
    config = get_example_config("1d")
    workflow = Workflow(config)
    region = CircleSkyRegion(center=SkyCoord("85d 23d"), radius=1 * u.deg)
    geom = WcsGeom.create(npix=(150, 150), binsz=0.05, skydir=SkyCoord("83d 22d"))
    exclusion_mask = ~geom.region_mask([region])

    filename = tmp_path / "exclusion.fits"
    exclusion_mask.write(filename)
    config.datasets.background.method = "reflected"
    config.datasets.background.exclusion = filename
    workflow.get_observations()
    workflow.get_datasets()
    assert len(workflow.datasets) == 2

    config = get_example_config("3d")
    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    geom = workflow.datasets[0]._geom
    exclusion_mask = ~geom.region_mask([region])
    filename = tmp_path / "exclusion3d.fits"
    exclusion_mask.write(filename)
    config.datasets.background.exclusion = filename
    workflow.get_datasets()
    assert len(workflow.datasets) == 1


@requires_data()
def test_workflow_1d_stacked_no_fit_range():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_cone: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 5 deg}
        obs_ids: [23592, 23559]

    datasets:
        type: 1d
        stack: false
        geom:
            axes:
                energy: {min: 0.01 TeV, max: 100 TeV, nbins: 73}
                energy_true: {min: 0.03 TeV, max: 100 TeV, nbins: 50}
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.1 deg}
        containment_correction: true
        background:
            method: reflected
    """
    config = WorkflowConfig.from_yaml(cfg)
    workflow = Workflow(config)
    workflow.update_config(cfg)
    workflow.config.datasets.stack = True
    workflow.get_observations()
    workflow.get_datasets()
    workflow.read_models(MODEL_FILE_1D)
    workflow.run_fit()
    with pytest.raises(ValueError):
        workflow.get_excess_map()

    assert len(workflow.datasets) == 1
    assert_allclose(workflow.datasets["stacked"].counts.data.sum(), 184)
    pars = workflow.models.parameters
    assert_allclose(workflow.datasets[0].mask_fit.data, True)

    assert_allclose(pars["index"].value, 2.76913, rtol=1e-2)
    assert_allclose(pars["amplitude"].value, 5.479729e-11, rtol=1e-2)


@requires_data()
def test_workflow_ring_background():
    config = get_example_config("3d")
    config.datasets.background.method = "ring"
    config.datasets.background.parameters = {"r_in": "0.7 deg", "width": "0.7 deg"}
    config.datasets.geom.axes.energy.nbins = 1
    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    workflow.get_excess_map()
    assert isinstance(workflow.datasets[0], MapDataset)
    assert_allclose(
        workflow.datasets[0].npred_background().data[0, 10, 10], 0.091799, rtol=1e-2
    )
    assert isinstance(workflow.excess_map["sqrt_ts"], WcsNDMap)
    assert_allclose(workflow.excess_map.npred_excess.data[0, 62, 62], 134.12389)


@requires_data()
def test_workflow_ring_3d():
    config = get_example_config("3d")
    config.datasets.background.method = "ring"
    config.datasets.background.parameters = {"r_in": "0.7 deg", "width": "0.7 deg"}
    workflow = Workflow(config)
    workflow.get_observations()
    with pytest.raises(ValueError):
        workflow.get_datasets()


@requires_data()
def test_workflow_no_bkg_1d(caplog):
    config = get_example_config("1d")
    workflow = Workflow(config)
    with caplog.at_level(logging.WARNING):
        workflow.get_observations()
        workflow.get_datasets()
        assert not isinstance(workflow.datasets[0], SpectrumDatasetOnOff)
        assert "No background maker set. Check configuration." in [
            _.message for _ in caplog.records
        ]


@requires_data()
def test_workflow_no_bkg_3d(caplog):
    config = get_example_config("3d")
    config.datasets.background.method = None
    workflow = Workflow(config)
    with caplog.at_level(logging.WARNING):
        workflow.get_observations()
        workflow.get_datasets()
        assert isinstance(workflow.datasets[0], MapDataset)
        assert "No background maker set. Check configuration." in [
            _.message for _ in caplog.records
        ]


@requires_data()
def test_workflow_3d():
    config = get_example_config("3d")
    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    workflow.read_models(MODEL_FILE)
    workflow.datasets["stacked"].background_model.spectral_model.tilt.frozen = False
    workflow.run_fit()
    workflow.get_flux_points()

    assert len(workflow.datasets) == 1
    assert len(workflow.models.parameters) == 8
    res = workflow.models.parameters
    assert res["amplitude"].unit == "cm-2 s-1 TeV-1"

    table = workflow.flux_points.data.to_table(sed_type="dnde")
    assert len(table) == 2
    dnde = table["dnde"].quantity

    assert_allclose(dnde[0].value, 1.2722e-11, rtol=1e-2)
    assert_allclose(dnde[-1].value, 4.054128e-13, rtol=1e-2)
    assert_allclose(res["index"].value, 2.772814, rtol=1e-2)
    assert_allclose(res["tilt"].value, -0.133436, rtol=1e-2)


@requires_data()
def test_workflow_3d_joint_datasets():
    config = get_example_config("3d")
    config.datasets.stack = False
    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    assert len(workflow.datasets) == 2

    assert_allclose(
        workflow.datasets[0].background_model.spectral_model.norm.value,
        1.031743694988066,
        rtol=1e-6,
    )
    assert_allclose(
        workflow.datasets[0].background_model.spectral_model.tilt.value,
        0.0,
        rtol=1e-6,
    )
    assert_allclose(
        workflow.datasets[1].background_model.spectral_model.norm.value,
        0.9776349021876344,
        rtol=1e-6,
    )


@requires_data()
def test_usage_errors():
    config = get_example_config("1d")
    workflow = Workflow(config)
    with pytest.raises(RuntimeError):
        workflow.get_datasets()
    with pytest.raises(RuntimeError):
        workflow.read_datasets()
    with pytest.raises(RuntimeError):
        workflow.write_datasets()
    with pytest.raises(TypeError):
        workflow.read_models()
    with pytest.raises(RuntimeError):
        workflow.write_models()
    with pytest.raises(RuntimeError):
        workflow.run_fit()
    with pytest.raises(RuntimeError):
        workflow.get_flux_points()
    with pytest.raises(ValidationError):
        workflow.config.datasets.type = "None"


@requires_data()
def test_datasets_io(tmpdir):
    config = get_example_config("3d")

    workflow = Workflow(config)
    workflow.get_observations()
    workflow.get_datasets()
    models_str = Path(MODEL_FILE).read_text()
    workflow.models = models_str

    config.general.datasets_file = tmpdir / "datasets.yaml"
    config.general.models_file = tmpdir / "models.yaml"
    workflow.write_datasets()
    workflow = Workflow(config)
    workflow.read_datasets()
    assert len(workflow.datasets.models) == 2
    assert workflow.models.names == ["source", "stacked-bkg"]

    workflow.models[0].parameters["index"].value = 3
    workflow.write_models()
    workflow = Workflow(config)
    workflow.read_datasets()
    assert len(workflow.datasets.models) == 2
    assert workflow.models.names == ["source", "stacked-bkg"]
    assert workflow.models[0].parameters["index"].value == 3



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .core import WcsMap
from .geom import WcsGeom
from .ndmap import WcsNDMap

__all__ = [
    "WcsGeom",
    "WcsMap",
    "WcsNDMap",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import operator
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord, angular_separation
from astropy.time import Time
from regions import CircleSkyRegion
from gammapy.data.gti import GTI
from gammapy.datasets.map import MapEvaluator
from gammapy.irf import EDispKernel, PSFKernel
from gammapy.maps import Map, MapAxis, RegionGeom, RegionNDMap, TimeMapAxis, WcsGeom
from gammapy.modeling import Parameter
from gammapy.modeling.models import (
    CompoundSpectralModel,
    ConstantSpatialModel,
    ConstantSpectralModel,
    ConstantTemporalModel,
    FoVBackgroundModel,
    GaussianSpatialModel,
    LightCurveTemplateTemporalModel,
    LogParabolaSpectralModel,
    Models,
    PiecewiseNormSpatialModel,
    PointSpatialModel,
    PowerLawNormSpectralModel,
    PowerLawSpectralModel,
    PowerLawTemporalModel,
    SkyModel,
    SpatialModel,
    TemplateNPredModel,
    TemplateSpatialModel,
    create_fermi_isotropic_diffuse_model,
)
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import mpl_plot_check, requires_data


@pytest.fixture(scope="session")
def sky_model():
    spatial_model = GaussianSpatialModel(
        lon_0="3 deg", lat_0="4 deg", sigma="3 deg", frame="galactic"
    )
    spectral_model = PowerLawSpectralModel(
        index=2, amplitude="1e-11 cm-2 s-1 TeV-1", reference="1 TeV"
    )
    spectral_model.index.error = 0.1
    spectral_model.amplitude.error = "1e-12 cm-2 s-1 TeV-1"

    temporal_model = ConstantTemporalModel()
    return SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model,
        temporal_model=temporal_model,
        name="source-1",
    )


@pytest.fixture(scope="session")
def gti():
    start = [1, 3, 5] * u.day
    stop = [2, 3.5, 6] * u.day
    t_ref = Time(55555, format="mjd")
    gti = GTI.create(start, stop, reference_time=t_ref)
    return gti


@pytest.fixture(scope="session")
def diffuse_model():
    axis = MapAxis.from_edges(
        [0.1, 1, 100], name="energy_true", unit="TeV", interp="log"
    )
    m = Map.create(
        npix=(4, 3), binsz=2, axes=[axis], unit="cm-2 s-1 MeV-1 sr-1", frame="galactic"
    )
    m.data += 42
    spatial_model = TemplateSpatialModel(
        m, normalize=False, filename="diffuse_test.fits"
    )
    return SkyModel(PowerLawNormSpectralModel(), spatial_model)


@pytest.fixture(scope="session")
def geom():
    axis = MapAxis.from_edges(np.logspace(-1, 1, 3), unit=u.TeV, name="energy")
    return WcsGeom.create(skydir=(0, 0), npix=(5, 4), frame="galactic", axes=[axis])


@pytest.fixture(scope="session")
def geom_true():
    axis = MapAxis.from_edges(np.logspace(-1, 1, 4), unit=u.TeV, name="energy_true")
    return WcsGeom.create(skydir=(0, 0), npix=(5, 4), frame="galactic", axes=[axis])


@pytest.fixture(scope="session")
def exposure(geom_true):
    m = Map.from_geom(geom_true)
    m.quantity = np.ones(geom_true.data_shape) * u.Quantity("100 m2 s")
    m.data[1] *= 10
    return m


@pytest.fixture(scope="session")
def background(geom):
    m = Map.from_geom(geom)
    m.quantity = np.ones(geom.data_shape) * 1e-7
    return m


@pytest.fixture(scope="session")
def edisp(geom, geom_true):
    e_reco = geom.axes["energy"]
    e_true = geom_true.axes["energy_true"]
    return EDispKernel.from_diagonal_response(
        energy_axis_true=e_true, energy_axis=e_reco
    )


@pytest.fixture(scope="session")
def psf(geom_true):
    sigma = 0.5 * u.deg
    return PSFKernel.from_gauss(geom_true, sigma)


@pytest.fixture(scope="session")
def evaluator(sky_model, exposure, psf, edisp, gti):
    return MapEvaluator(sky_model, exposure, psf=psf, edisp=edisp, gti=gti)


@pytest.fixture(scope="session")
def diffuse_evaluator(diffuse_model, exposure, psf, edisp):
    return MapEvaluator(diffuse_model, exposure, psf=psf, edisp=edisp)


@pytest.fixture(scope="session")
def diffuse_evaluator_edisp_false(diffuse_model, exposure, psf, edisp):
    model = diffuse_model.copy()
    model.apply_irf["edisp"] = False
    return MapEvaluator(model, exposure, psf=psf, edisp=edisp)


@pytest.fixture(scope="session")
def sky_models(sky_model):
    sky_model_2 = sky_model.copy(name="source-2")
    sky_model_3 = sky_model.copy(name="source-3")
    return Models([sky_model_2, sky_model_3])


@pytest.fixture(scope="session")
def sky_models_2(sky_model):
    sky_model_4 = sky_model.copy(name="source-4")
    sky_model_5 = sky_model.copy(name="source-5")
    return Models([sky_model_4, sky_model_5])


@requires_data()
def test_sky_model_init():
    with pytest.raises(TypeError):
        spatial_model = GaussianSpatialModel()
        SkyModel(spectral_model=1234, spatial_model=spatial_model)

    with pytest.raises(TypeError):
        SkyModel(spectral_model=PowerLawSpectralModel(), spatial_model=1234)

    # test init of energy dependent temporal models
    filename = make_path(
        "$GAMMAPY_DATA/gravitational_waves/GW_example_DC_map_file.fits.gz"
    )
    temporal_model = LightCurveTemplateTemporalModel.read(filename, format="map")
    spatial_model = PointSpatialModel()
    spectral_model_fake = ConstantSpectralModel()

    model = SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model_fake,
        temporal_model=temporal_model,
        name="test-source",
    )
    assert model.name == "test-source"


def test_sky_model_spatial_none_io(tmpdir):
    pwl = PowerLawSpectralModel()
    model = SkyModel(spectral_model=pwl, name="test")
    models = Models([model])

    filename = tmpdir / "test-models-none.yaml"
    models.write(filename)

    models = Models.read(filename)

    assert models["test"].spatial_model is None


def test_sky_model_spatial_none_evaluate(geom_true, gti):
    pwl = PowerLawSpectralModel()
    model = SkyModel(spectral_model=pwl, name="test")

    data = model.evaluate_geom(geom_true, gti).to_value("cm-2 s-1 TeV-1")

    assert data.shape == (3, 1, 1)
    assert_allclose(data[0], 1.256774e-11, rtol=1e-6)


def test_skymodel_addition(sky_model, sky_models, sky_models_2, diffuse_model):
    models = sky_model + sky_model.copy()
    assert isinstance(models, Models)
    assert len(models) == 2

    models = sky_model + sky_models
    assert isinstance(models, Models)
    assert len(models) == 3

    models = sky_models + sky_model
    assert isinstance(models, Models)
    assert len(models) == 3

    models = sky_models + diffuse_model
    assert isinstance(models, Models)
    assert len(models) == 3

    models = sky_models + sky_models_2
    assert isinstance(models, Models)
    assert len(models) == 4

    models = sky_model + sky_models
    assert isinstance(models, Models)
    assert len(models) == 3


def test_background_model(background):
    bkg1 = TemplateNPredModel(background)
    bkg1.spectral_model.norm.value = 2.0
    npred1 = bkg1.evaluate()
    assert_allclose(npred1.data[0][0][0], background.data[0][0][0] * 2.0, rtol=1e-3)
    assert_allclose(npred1.data.sum(), background.data.sum() * 2.0, rtol=1e-3)

    bkg2 = TemplateNPredModel(background)
    bkg2.spectral_model.norm.value = 2.0
    bkg2.spectral_model.tilt.value = 0.2
    bkg2.spectral_model.reference.quantity = "1000 GeV"

    npred2 = bkg2.evaluate()
    assert_allclose(npred2.data[0][0][0], 2.254e-07, rtol=1e-3)
    assert_allclose(npred2.data.sum(), 7.352e-06, rtol=1e-3)


def test_background_slice(background):
    bkg1 = TemplateNPredModel(background)
    e_edges = background.geom.axes[0].edges
    bkg1_slice = bkg1.slice_by_energy(e_edges[0], e_edges[1])  # 1 bin slice
    assert bkg1_slice.name == bkg1_slice.name
    assert bkg1_slice.map.data.shape == bkg1.map.sum_over_axes().data.shape
    assert_allclose(bkg1_slice.map.data[0, :, :], bkg1.map.data[0, :, :], rtol=1e-5)


def test_background_model_io(tmpdir, background):
    filename = str(tmpdir / "test-bkg-file.fits")
    bkg = TemplateNPredModel(background, filename=filename)
    bkg.spectral_model.norm.value = 2.0
    bkg.write(overwrite=False)
    bkg.write(overwrite=True)
    bkg_dict = bkg.to_dict()
    bkg_read = bkg.from_dict(bkg_dict)

    assert_allclose(
        bkg_read.evaluate().data.sum(), background.data.sum() * 2.0, rtol=1e-3
    )
    assert bkg_read.filename == filename


def test_background_model_io_missing_file(tmpdir, background):
    bkg = TemplateNPredModel(background, filename=None)
    with pytest.raises(IOError):
        bkg.write(overwrite=True)


def test_background_model_copy(background):
    background_copy = background.copy()
    bkg = TemplateNPredModel(background_copy)
    bkg.map.data += 1.0
    assert np.all(
        background_copy.data == background.data
    )  # Check that the original map is unchanged

    bkg_copy = bkg.copy()
    bkg_copy.map.data += 1.0
    assert np.all(
        bkg_copy.map.data == bkg.map.data
    )  # Check that the map has now changed


def test_parameters(sky_models):
    parnames = [
        "index",
        "amplitude",
        "reference",
        "lon_0",
        "lat_0",
        "sigma",
        "e",
        "phi",
    ] * 2
    assert sky_models.parameters.names == parnames

    # Check that model parameters are references to the parts
    p1 = sky_models.parameters["lon_0"]
    p2 = sky_models[0].parameters["lon_0"]
    assert p1 is p2


def test_str(sky_models):
    assert "Component 0" in str(sky_models)
    assert "Component 1" in str(sky_models)


def test_get_item(sky_models):
    model = sky_models["source-2"]
    assert model.name == "source-2"

    model = sky_models["source-3"]
    assert model.name == "source-3"

    with pytest.raises(ValueError):
        sky_models["spam"]


def test_names(sky_models):
    assert sky_models.names == ["source-2", "source-3"]


@requires_data()
def test_models_mutation(sky_model, sky_models, sky_models_2):
    mods = sky_models

    mods.insert(0, sky_model)
    assert mods.names == ["source-1", "source-2", "source-3"]

    mods.extend(sky_models_2)
    assert mods.names == ["source-1", "source-2", "source-3", "source-4", "source-5"]

    mod3 = mods[3]
    mods.remove(mods[3])
    assert mods.names == ["source-1", "source-2", "source-3", "source-5"]
    mods.append(mod3)
    assert mods.names == ["source-1", "source-2", "source-3", "source-5", "source-4"]
    mods.pop(3)
    assert mods.names == ["source-1", "source-2", "source-3", "source-4"]

    with pytest.raises(ValueError, match="Model names must be unique"):
        mods.append(sky_model)
    with pytest.raises(ValueError, match="Model names must be unique"):
        mods.insert(0, sky_model)
    with pytest.raises(ValueError, match="Model names must be unique"):
        mods.extend(sky_models_2)
    with pytest.raises(ValueError, match="Model names must be unique"):
        mods = sky_models + sky_models_2
    with pytest.raises(ValueError, match="Model names must be unique"):
        mods[1] = mods[0]

    mods[1] = mods[1]
    assert mods.names == ["source-1", "source-2", "source-3", "source-4"]
    mods[1] = mods[1].copy(name="copy")
    assert mods.names == ["source-1", "copy", "source-3", "source-4"]


class TestSkyModel:
    @staticmethod
    def test_repr(sky_model):
        assert "SkyModel" in repr(sky_model)

    @staticmethod
    def test_str(sky_model):
        string_model = str(sky_model)
        model_lines = string_model.splitlines()
        assert "SkyModel" in string_model
        assert "2.000   +/-    0.10" in model_lines[8]

    @staticmethod
    def test_parameters(sky_model):
        # Check that model parameters are references to the spatial and spectral parts
        p1 = sky_model.parameters["lon_0"]
        p2 = sky_model.spatial_model.parameters["lon_0"]
        assert p1 is p2

        p1 = sky_model.parameters["amplitude"]
        p2 = sky_model.spectral_model.parameters["amplitude"]
        assert p1 is p2

    @staticmethod
    def test_evaluate_scalar(sky_model):
        lon = 3 * u.deg
        lat = 4 * u.deg
        energy = 1 * u.TeV

        q = sky_model.evaluate(lon, lat, energy)

        assert q.unit == "cm-2 s-1 TeV-1 sr-1"
        assert np.isscalar(q.value)
        assert_allclose(q.to_value("cm-2 s-1 TeV-1 deg-2"), 1.76879232e-13)

    @staticmethod
    def test_evaluate_array(sky_model):
        lon = 3 * u.deg * np.ones(shape=(3, 4))
        lat = 4 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = sky_model.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.shape == (5, 3, 4)
        assert_allclose(q.to_value("cm-2 s-1 TeV-1 deg-2"), 1.76879232e-13)

    @staticmethod
    def test_processing(sky_model):
        assert sky_model.apply_irf == {"exposure": True, "psf": True, "edisp": True}
        out = sky_model.to_dict()
        assert "apply_irf" not in out

        sky_model.apply_irf["edisp"] = False
        out = sky_model.to_dict()
        assert out["apply_irf"] == {"exposure": True, "psf": True, "edisp": False}
        sky_model.apply_irf["edisp"] = True


class Test_Template_with_cube:
    @staticmethod
    def test_evaluate_scalar(diffuse_model):
        # Check pixel inside map
        val = diffuse_model.evaluate(0 * u.deg, 0 * u.deg, 10 * u.TeV)
        assert val.unit == "cm-2 s-1 MeV-1 sr-1"
        assert val.shape == (1,)
        assert_allclose(val.value, 42)

        # Check pixel outside map (spatially)
        val = diffuse_model.evaluate(100 * u.deg, 0 * u.deg, 10 * u.TeV)
        assert_allclose(val.value, 0)

        # Check pixel outside energy range
        val = diffuse_model.evaluate(0 * u.deg, 0 * u.deg, 200 * u.TeV)
        assert_allclose(val.value, 0)

    @staticmethod
    def test_evaluate_array(diffuse_model):
        lon = 1 * u.deg * np.ones(shape=(3, 4))
        lat = 2 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = diffuse_model.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.shape == (5, 3, 4)
        assert_allclose(q.value.mean(), 42)

    @staticmethod
    def test_write(tmpdir, diffuse_model):
        filename = tmpdir / diffuse_model.spatial_model.filename

        diffuse_model.spatial_model.filename = None
        with pytest.raises(IOError):
            diffuse_model.spatial_model.write()

        with pytest.raises(IOError):
            Models(diffuse_model).to_dict()

        diffuse_model.spatial_model.filename = filename
        diffuse_model.spatial_model.write(overwrite=False)
        TemplateSpatialModel.read(filename)

    @staticmethod
    @requires_data()
    def test_read():
        model = TemplateSpatialModel.read(
            "$GAMMAPY_DATA/tests/unbundled/fermi/gll_iem_v02_cutout.fits",
            normalize=False,
        )
        assert model.map.unit == "cm-2 s-1 MeV-1 sr-1"

        # Check pixel inside map
        val = model.evaluate(0 * u.deg, 0 * u.deg, energy=100 * u.GeV)
        assert val.unit == "cm-2 s-1 MeV-1 sr-1"
        assert val.shape == (1,)
        assert_allclose(val.value, 1.395156e-12, rtol=1e-5)

    @staticmethod
    def test_evaluation_radius(diffuse_model):
        radius = diffuse_model.evaluation_radius
        assert radius.unit == "deg"
        assert_allclose(radius.value, 4)

    @staticmethod
    def test_frame(diffuse_model):
        assert diffuse_model.frame == "galactic"

    @staticmethod
    def test_processing(diffuse_model):
        assert diffuse_model.apply_irf == {"exposure": True, "psf": True, "edisp": True}
        out = diffuse_model.to_dict()
        assert "apply_irf" not in out

        diffuse_model.apply_irf["edisp"] = False
        out = diffuse_model.to_dict()
        assert out["apply_irf"] == {"exposure": True, "psf": True, "edisp": False}
        diffuse_model.apply_irf["edisp"] = True

    @staticmethod
    def test_datasets_name(diffuse_model):
        assert diffuse_model.datasets_names is None

        diffuse_model.datasets_names = ["1", "2"]
        out = diffuse_model.to_dict()
        assert out["datasets_names"] == ["1", "2"]

        diffuse_model.datasets_names = None
        out = diffuse_model.to_dict()
        assert "datasets_names" not in out


class Test_template_cube_MapEvaluator:
    @staticmethod
    def test_compute_dnde(diffuse_evaluator):
        out = diffuse_evaluator.compute_dnde()
        assert out.shape == (3, 4, 5)
        out = out.to("cm-2 s-1 MeV-1 sr-1")
        assert_allclose(out.value.sum(), 2520.0, rtol=1e-5)
        assert_allclose(out.value[0, 0, 0], 42, rtol=1e-5)

    @staticmethod
    def test_compute_flux(diffuse_evaluator):
        out = diffuse_evaluator.compute_flux()
        assert out.data.shape == (3, 4, 5)
        out = out.quantity.to("cm-2 s-1")
        assert_allclose(out.value.sum(), 633263.444803, rtol=5e-3)
        assert_allclose(out.value[0, 0, 0], 1164.656176, rtol=5e-3)

    @staticmethod
    def test_apply_psf(diffuse_evaluator):
        flux = diffuse_evaluator.compute_flux()
        npred = diffuse_evaluator.apply_exposure(flux)
        out = diffuse_evaluator.apply_psf(npred)
        assert out.data.shape == (3, 4, 5)
        assert_allclose(out.data.sum(), 1.106404e12, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 5.586508e08, rtol=5e-3)

    @staticmethod
    def test_apply_edisp(diffuse_evaluator):
        flux = diffuse_evaluator.compute_flux()
        npred = diffuse_evaluator.apply_exposure(flux)
        out = diffuse_evaluator.apply_edisp(npred)
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 1.606345e12, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 1.83018e10, rtol=5e-3)

    @staticmethod
    def test_compute_npred(diffuse_evaluator):
        out = diffuse_evaluator.compute_npred()
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 1.106403e12, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 8.778828e09, rtol=5e-3)

    @staticmethod
    def test_apply_edisp_false(diffuse_evaluator_edisp_false):
        out = diffuse_evaluator_edisp_false.compute_npred()
        assert "energy" in out.geom.axes.names
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 1.106403e12, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 8.778828e09, rtol=5e-3)


class TestSkyModelMapEvaluator:
    @staticmethod
    def test_compute_dnde(evaluator):
        out = evaluator.compute_dnde()
        assert out.shape == (3, 4, 5)
        assert out.unit == "cm-2 s-1 TeV-1 sr-1"
        assert_allclose(
            out.to_value("cm-2 s-1 TeV-1 deg-2").sum(),
            1.1788166328203174e-11,
            rtol=1e-5,
        )
        assert_allclose(
            out.to_value("cm-2 s-1 TeV-1 deg-2")[0, 0, 0],
            5.087056282039508e-13,
            rtol=1e-5,
        )

    @staticmethod
    def test_compute_flux(evaluator):
        out = evaluator.compute_flux()
        out = out.quantity.to_value("cm-2 s-1")
        assert out.shape == (3, 4, 5)
        assert_allclose(out.sum(), 2.213817e-12, rtol=1e-5)
        assert_allclose(out[0, 0, 0], 7.938388e-14, rtol=1e-5)

    @staticmethod
    def test_apply_psf(evaluator):
        flux = evaluator.compute_flux()
        npred = evaluator.apply_exposure(flux)
        out = evaluator.apply_psf(npred)
        assert out.data.shape == (3, 4, 5)
        assert_allclose(out.data.sum(), 3.862314e-06, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 4.126612e-08, rtol=5e-3)

    @staticmethod
    def test_apply_edisp(evaluator):
        flux = evaluator.compute_flux()
        npred = evaluator.apply_exposure(flux)
        out = evaluator.apply_edisp(npred)
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 5.615601e-06, rtol=1e-5)
        assert_allclose(out.data[0, 0, 0], 1.33602e-07, rtol=1e-5)

    @staticmethod
    def test_compute_npred(evaluator, gti):
        out = evaluator.compute_npred()
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 3.862314e-06, rtol=5e-3)
        assert_allclose(out.data[0, 0, 0], 6.94503e-08, rtol=5e-3)


def test_sky_point_source():
    # Test special case of point source. Regression test for GH 2367.

    energy_axis = MapAxis.from_edges(
        [1, 10], unit="TeV", name="energy_true", interp="log"
    )
    exposure = Map.create(
        skydir=(100, 70),
        npix=(4, 4),
        binsz=0.1,
        proj="AIT",
        unit="cm2 s",
        axes=[energy_axis],
    )
    exposure.data = np.ones_like(exposure.data)

    spatial_model = PointSpatialModel(
        lon_0=100.06 * u.deg, lat_0=70.03 * u.deg, frame="icrs"
    )
    # Create a spectral model with integral flux of 1 cm-2 s-1 in this energy band
    spectral_model = ConstantSpectralModel(const="1 cm-2 s-1 TeV-1")
    spectral_model.const.value /= spectral_model.integral(1 * u.TeV, 10 * u.TeV).value
    model = SkyModel(spatial_model=spatial_model, spectral_model=spectral_model)
    evaluator = MapEvaluator(model=model, exposure=exposure)
    flux = evaluator.compute_flux().quantity.to_value("cm-2 s-1")[0]

    expected = [
        [0, 0, 0, 0],
        [0, 0.140, 0.058, 0.0],
        [0, 0.564, 0.236, 0],
        [0, 0, 0, 0],
    ]
    assert_allclose(flux, expected, atol=0.01)

    assert_allclose(flux.sum(), 1)


@requires_data()
def test_fermi_isotropic():
    filename = "$GAMMAPY_DATA/fermi_3fhl/iso_P8R2_SOURCE_V6_v06.txt"
    energy = [0.01, 1, 10, 100, 1000] * u.GeV
    coords = {"lon": 0 * u.deg, "lat": 0 * u.deg, "energy": energy}

    model_noextrapolate = create_fermi_isotropic_diffuse_model(
        filename=filename,
        interp_kwargs={"extrapolate": False},
    )
    model_extrapolate = create_fermi_isotropic_diffuse_model(
        filename=filename,
        interp_kwargs={"extrapolate": True, "method": "nearest"},
    )

    flux_noextrapolate = model_noextrapolate(**coords)
    assert_allclose(
        flux_noextrapolate.value,
        [np.nan, 5.98959823e-10, 6.26407059e-12, 2.83721193e-14, np.nan],
        rtol=1e-3,
    )
    assert flux_noextrapolate.unit == "MeV-1 cm-2 s-1 sr-1"
    assert isinstance(model_noextrapolate.spectral_model, CompoundSpectralModel)

    assert_allclose(
        model_extrapolate(**coords).value,
        [2.52894e-06, 5.86237e-10, 5.78221e-12, 2.32045e-14, 2.74918e-16],
        rtol=1e-3,
    )

    # No extrapolation with bounds_error
    with pytest.raises(ValueError):
        create_fermi_isotropic_diffuse_model(
            filename=filename,
            interp_kwargs={"extrapolate": False, "bounds_error": True},
        )


class MyCustomGaussianModel(SpatialModel):
    """My custom gaussian model.

    Parameters
    ----------
    lon_0, lat_0 : `~astropy.coordinates.Angle`
        Center position
    sigma_1TeV : `~astropy.coordinates.Angle`
        Width of the Gaussian at 1 TeV
    sigma_10TeV : `~astropy.coordinates.Angle`
        Width of the Gaussian at 10 TeV

    """

    tag = "MyCustomGaussianModel"
    lon_0 = Parameter("lon_0", "0 deg")
    lat_0 = Parameter("lat_0", "0 deg", min=-90, max=90)

    sigma_1TeV = Parameter("sigma_1TeV", "0.5 deg", min=0)
    sigma_10TeV = Parameter("sigma_10TeV", "0.1 deg", min=0)

    @staticmethod
    def evaluate(lon, lat, energy, lon_0, lat_0, sigma_1TeV, sigma_10TeV):
        """Evaluate custom Gaussian model"""
        sigmas = u.Quantity([sigma_1TeV, sigma_10TeV])
        energy_nodes = [1, 10] * u.TeV
        sigma = np.interp(energy, energy_nodes, sigmas)
        sigma = sigma.to("rad")

        sep = angular_separation(lon, lat, lon_0, lat_0)

        exponent = -0.5 * (sep / sigma) ** 2
        norm = 1 / (2 * np.pi * sigma**2)
        return norm * np.exp(exponent)

    @property
    def evaluation_radius(self):
        """Evaluation radius (`~astropy.coordinates.Angle`)."""
        return 5 * self.sigma_1TeV.quantity


def test_energy_dependent_model():
    axis = MapAxis.from_edges(np.logspace(-1, 1, 4), unit=u.TeV, name="energy_true")
    geom_true = WcsGeom.create(
        skydir=(0, 0), binsz="0.1 deg", npix=(50, 50), frame="galactic", axes=[axis]
    )

    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    spatial_model = MyCustomGaussianModel(frame="galactic")
    sky_model = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)
    model = sky_model.integrate_geom(geom_true)

    assert_allclose(model.data.sum(), 9.9e-11, rtol=1e-3)


def test_plot_grid(geom_true):
    spatial_model = MyCustomGaussianModel(frame="galactic")
    with mpl_plot_check():
        spatial_model.plot_grid(geom=geom_true)


def test_sky_model_create():
    m = SkyModel.create("pl", "point", name="my-source")
    assert isinstance(m.spatial_model, PointSpatialModel)
    assert isinstance(m.spectral_model, PowerLawSpectralModel)
    assert m.name == "my-source"


def test_integrate_geom():
    model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )
    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    sky_model = SkyModel(spectral_model=spectral_model, spatial_model=model)

    center = SkyCoord("0d", "0d", frame="icrs")
    radius = 0.3 * u.deg
    square = CircleSkyRegion(center, radius)

    axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3, name="energy_true")
    geom = RegionGeom(region=square, axes=[axis], binsz_wcs="0.01deg")

    integral = sky_model.integrate_geom(geom).data

    assert_allclose(integral / 1e-12, [[[5.299]], [[2.460]], [[1.142]]], rtol=1e-3)


def test_evaluate_integrate_nd_geom():
    model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )
    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    sky_model = SkyModel(spectral_model=spectral_model, spatial_model=model)

    center = SkyCoord("0d", "0d", frame="icrs")
    radius = 0.3 * u.deg
    region = CircleSkyRegion(center, radius)

    energy_axis = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=3, name="energy_true"
    )
    other_axis = MapAxis.from_edges([0.0, 1.0, 2.0], name="other")

    wcs_geom = WcsGeom.create(
        width=[1, 1.2], binsz=0.05, skydir=center, axes=[energy_axis, other_axis]
    )
    region_geom = RegionGeom(
        region=region, axes=[other_axis, energy_axis], binsz_wcs="0.01deg"
    )

    evaluation = sky_model.evaluate_geom(wcs_geom)
    assert evaluation.shape == (2, 3, 24, 20)
    assert_allclose(evaluation[0], evaluation[1])
    assert_allclose(
        evaluation.value[0, :, 12, 10],
        [2.278184e-07, 4.908198e-08, 1.057439e-08],
        rtol=1e-6,
    )

    integral = sky_model.integrate_geom(wcs_geom).data
    assert integral.shape == (2, 3, 24, 20)
    assert_allclose(integral[0], integral[1])
    assert_allclose(integral[0, :, 12, 10], [1.973745e-13, 9.161312e-14, 4.252304e-14])

    integral = sky_model.integrate_geom(region_geom).data

    assert integral.shape == (3, 2, 1, 1)
    assert_allclose(integral[:, 0, :, :], integral[:, 1, :, :])
    assert_allclose(
        integral[:, 0] / 1e-12, [[[5.299]], [[2.460]], [[1.142]]], rtol=1e-3
    )


def test_evaluate_integrate_geom_with_time():
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )
    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    temporal_model = PowerLawTemporalModel()
    temporal_model.t_ref.value = 55000
    sky_model = SkyModel(
        spectral_model=spectral_model,
        spatial_model=spatial_model,
        temporal_model=temporal_model,
    )

    center = SkyCoord("0d", "0d", frame="icrs")
    energy_axis = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=3, name="energy_true"
    )
    other_axis = MapAxis.from_edges([0.0, 1.0, 2.0], name="other")

    t_ref = Time(temporal_model.t_ref.value, format="mjd")
    time_min = t_ref + [1, 3, 5, 7] * u.day
    time_max = t_ref + [2, 4, 6, 8] * u.day
    time_axis = TimeMapAxis.from_time_edges(time_min=time_min, time_max=time_max)

    wcs_geom = WcsGeom.create(
        width=[1, 1.2],
        binsz=0.05,
        skydir=center,
        axes=[energy_axis, other_axis, time_axis],
    )
    unit_exp = 1 / u.TeV / u.cm**2 / u.s / u.sr

    evaluation = sky_model.evaluate_geom(wcs_geom)
    assert evaluation.shape == (4, 2, 3, 24, 20)
    assert evaluation.unit.is_equivalent(unit_exp)
    assert_allclose(
        evaluation.value[0, 0, 1, 12, 10],
        7.362297e-08,
        rtol=1e-6,
    )

    radius = 0.3 * u.deg
    region = CircleSkyRegion(center, radius)

    sky_model1 = SkyModel(spectral_model=spectral_model, temporal_model=temporal_model)
    region_geom = RegionGeom(
        region=region, axes=[time_axis, energy_axis], binsz_wcs="0.01deg"
    )
    with pytest.raises(ValueError):
        sky_model1.evaluate_geom(region_geom)

    region_geom = RegionGeom(
        region=region, axes=[energy_axis, time_axis], binsz_wcs="0.01deg"
    )
    evaluation = sky_model1.evaluate_geom(geom=region_geom)
    assert evaluation.shape == (4, 3, 1, 1)
    assert_allclose(
        evaluation[0].value,
        [[[6.96238325e-12]], [[1.5000000e-12]], [[3.23165204e-13]]],
        rtol=1e-6,
    )
    unit_exp = 1 / u.TeV / u.cm**2 / u.s
    assert evaluation.unit.is_equivalent(unit_exp)

    integral = sky_model1.integrate_geom(geom=region_geom)
    assert integral.geom.data_shape == (4, 3, 1, 1)
    assert_allclose(
        integral.data[0],
        [[[8.03761675e-12]], [[3.73073122e-12]], [[1.73165204e-12]]],
        rtol=1e-6,
    )
    unit_exp = 1 / u.cm**2 / u.s
    assert integral.unit.is_equivalent(unit_exp)


def test_evaluate_integrate_geom_with_time_and_gti():
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )
    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    temporal_model = PowerLawTemporalModel()
    temporal_model.t_ref.value = 55000
    sky_model = SkyModel(
        spectral_model=spectral_model,
        spatial_model=spatial_model,
        temporal_model=temporal_model,
    )

    center = SkyCoord("0d", "0d", frame="icrs")

    start = np.linspace(0, 8, 10) * u.day
    stop = np.linspace(0.5, 8.5, 10) * u.day
    t_ref = Time(temporal_model.t_ref.value, format="mjd")
    gti = GTI.create(start, stop, reference_time=t_ref)

    energy_axis = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=3, name="energy_true"
    )
    other_axis = MapAxis.from_edges([0.0, 1.0, 2.0], name="other")

    time_min = t_ref + [1, 3, 5, 7] * u.day
    time_max = t_ref + [2, 4, 6, 8] * u.day

    time_axis = TimeMapAxis.from_time_edges(time_min=time_min, time_max=time_max)

    wcs_geom = WcsGeom.create(
        width=[1, 1.2],
        binsz=0.05,
        skydir=center,
        axes=[energy_axis, other_axis, time_axis],
    )

    evaluation = sky_model.evaluate_geom(geom=wcs_geom, gti=gti)
    assert evaluation.shape == (4, 2, 3, 24, 20)
    unit_exp = 1 / u.TeV / u.cm**2 / u.s / u.sr
    assert evaluation.unit.is_equivalent(unit_exp)
    assert_allclose(
        evaluation.value[0, 0, 1, 12, 10],
        7.102014e-08,
        rtol=1e-6,
    )

    radius = 0.3 * u.deg
    region = CircleSkyRegion(center, radius)

    sky_model1 = SkyModel(spectral_model=spectral_model, temporal_model=temporal_model)
    region_geom = RegionGeom(
        region=region, axes=[time_axis, energy_axis], binsz_wcs="0.01deg"
    )
    with pytest.raises(ValueError):
        sky_model1.evaluate_geom(region_geom, gti)

    region_geom = RegionGeom(
        region=region, axes=[energy_axis, time_axis], binsz_wcs="0.01deg"
    )
    evaluation = sky_model1.evaluate_geom(geom=region_geom, gti=gti)
    assert evaluation.shape == (4, 3, 1, 1)
    assert_allclose(
        evaluation[0].value,
        [[[6.71623839e-12]], [[1.44696970e-12]], [[3.11740171e-13]]],
        rtol=1e-3,
    )
    unit_exp = 1 / u.TeV / u.cm**2 / u.s
    assert evaluation.unit.is_equivalent(unit_exp)

    integral = sky_model1.integrate_geom(geom=region_geom, gti=gti)
    assert integral.geom.data_shape == (4, 3, 1, 1)
    assert_allclose(
        integral.data[0],
        [[[7.75345858e-12]], [[3.59883668e-12]], [[1.67043201e-12]]],
        rtol=1e-3,
    )
    unit_exp = 1 / u.cm**2 / u.s
    assert integral.unit.is_equivalent(unit_exp)


def test_compound_spectral_model(caplog):
    spatial_model = GaussianSpatialModel(
        lon_0="3 deg", lat_0="4 deg", sigma="3 deg", frame="galactic"
    )
    pwl = PowerLawSpectralModel(
        index=2, amplitude="1e-11 cm-2 s-1 TeV-1", reference="1 TeV"
    )
    lp = LogParabolaSpectralModel(
        amplitude="1e-12 cm-2 s-1 TeV-1", reference="10 TeV", alpha=2.0, beta=1.0
    )
    temporal_model = ConstantTemporalModel()

    spectral_model = CompoundSpectralModel(pwl, lp, operator.add)
    m = SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model,
        temporal_model=temporal_model,
        name="source-1",
    )
    assert_allclose(m.spectral_model(5 * u.TeV).value, 2.87e-12, rtol=1e-2)


def test_sky_model_contributes_point_region():
    model = SkyModel.create("pl", "point")

    geom = RegionGeom.create("icrs;point(0.05, 0.05)", binsz_wcs="0.01 deg")
    mask = RegionNDMap.from_geom(geom)
    assert np.any(model.contributes(mask))


def test_spatial_model_background(background):
    geom = background.geom

    spatial_model = ConstantSpatialModel(frame="galactic")
    identical_npred = TemplateNPredModel(
        background, spatial_model=spatial_model
    ).evaluate()
    assert_allclose(identical_npred, background.data)

    reference = FoVBackgroundModel(
        spatial_model=None, dataset_name="test"
    ).evaluate_geom(geom)
    identical = FoVBackgroundModel(
        spatial_model=spatial_model, dataset_name="test"
    ).evaluate_geom(geom)
    assert_allclose(identical, reference)

    spatial_model2 = ConstantSpatialModel(frame="galactic")
    spatial_model2.value.value = 2
    twice_npred = TemplateNPredModel(
        background, spatial_model=spatial_model2
    ).evaluate()
    assert_allclose(twice_npred, background.data * 2)

    twice = FoVBackgroundModel(
        spatial_model=spatial_model2, dataset_name="test"
    ).evaluate_geom(geom)
    assert_allclose(twice, reference * 2)


def test_spatial_model_io_background(tmp_path, background):
    spatial_model = ConstantSpatialModel(frame="galactic")

    fbkg_irf = str(tmp_path / "background_irf_test.fits")

    model = TemplateNPredModel(background, spatial_model=None, filename=fbkg_irf)
    model.write()

    model_dict = model.to_dict()
    assert "spatial" not in model_dict
    new_model = TemplateNPredModel.from_dict(model_dict)
    assert new_model.spatial_model is None

    model = TemplateNPredModel(
        background, spatial_model=spatial_model, filename=fbkg_irf
    )
    model_dict = model.to_dict()
    assert "spatial" in model_dict
    new_model = TemplateNPredModel.from_dict(model_dict)
    assert isinstance(new_model.spatial_model, ConstantSpatialModel)
    assert new_model.spatial_model.frame == "icrs"

    model = FoVBackgroundModel(spatial_model=None, dataset_name="test")
    model_dict = model.to_dict()
    assert "spatial" not in model_dict
    new_model = FoVBackgroundModel.from_dict(model_dict)
    assert new_model.spatial_model is None

    model = FoVBackgroundModel(spatial_model=spatial_model, dataset_name="test")
    model_dict = model.to_dict()
    assert "spatial" in model_dict
    new_model = FoVBackgroundModel.from_dict(model_dict)
    assert isinstance(new_model.spatial_model, ConstantSpatialModel)


def test_piecewise_spatial_model_background(background):
    geom = background.geom
    coords = geom.to_image().get_coord().flat

    spatial_model = PiecewiseNormSpatialModel(coords, frame="galactic")
    identical_npred = TemplateNPredModel(
        background, spatial_model=spatial_model
    ).evaluate()
    assert_allclose(identical_npred, background.data)

    reference = Map.from_geom(geom, data=1)
    identical = FoVBackgroundModel(
        spatial_model=spatial_model, dataset_name="test"
    ).evaluate_geom(geom)
    assert_allclose(identical, reference)

    spatial_model2 = PiecewiseNormSpatialModel(
        coords, norms=2 * np.ones(coords.shape[0]), frame="galactic"
    )
    twice_npred = TemplateNPredModel(
        background, spatial_model=spatial_model2
    ).evaluate()
    assert_allclose(twice_npred, background.data * 2)

    twice = FoVBackgroundModel(
        spatial_model=spatial_model2, dataset_name="test"
    ).evaluate_geom(geom)
    assert_allclose(twice, reference * 2.0)

    copied = FoVBackgroundModel(spatial_model=spatial_model, dataset_name="test").copy()
    assert isinstance(copied.spatial_model, PiecewiseNormSpatialModel)

    assert "Spatial model type" in copied.__str__()


def test_naming_fov_background_model():
    fov_default = FoVBackgroundModel(dataset_name="one", name=None)
    assert fov_default.name == "one-bkg"
    fov_named = FoVBackgroundModel(dataset_name="two", name="custom_name")
    assert fov_named.name == "custom_name"



"""
Maps
====

A thorough tutorial to work with WCS maps.

.. figure:: ../../_static/gammapy_maps.png
   :alt: Gammapy Maps Illustration

   Gammapy Maps Illustration

Introduction
------------

The `gammapy.maps` submodule contains classes for representing
pixilised data on the sky with an arbitrary number of non-spatial
dimensions such as energy, time, event class or any possible
user-defined dimension (illustrated in the image above). The main
`~gammapy.maps.Map` data structure features a uniform API for
`WCS <https://fits.gsfc.nasa.gov/fits_wcs.html>`__ as well as
`HEALPix <https://en.wikipedia.org/wiki/HEALPix>`__ based images. The
API also generalizes simple image based operations such as smoothing,
interpolation and reprojection to the arbitrary extra dimensions and
makes working with (2 + N)-dimensional hypercubes as easy as working
with a simple 2D image. Further information is also provided on the
`~gammapy.maps` docs page.

In the following introduction we will learn all the basics of working
with WCS based maps. HEALPix based maps will be covered in a future
tutorial. Make sure you have worked through the :doc:`Gammapy
overview </tutorials/starting/overview>`, because a solid knowledge
about working with `~astropy.coordinates.SkyCoord` and `~astropy.units.Quantity` objects as well as
`Numpy <http://www.numpy.org/>`__ is required for this tutorial.

This notebook is rather lengthy, but getting to know the `~gammapy.maps.Map` data
structure in detail is essential for working with Gammapy and will allow
you to fulfill complex analysis tasks with very few and simple code in
future!

"""

######################################################################
# Setup
# -----
#

import os

# %matplotlib inline
import numpy as np
from astropy import units as u
from astropy.convolution import convolve
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.data import EventList
from gammapy.maps import (
    LabelMapAxis,
    Map,
    MapAxes,
    MapAxis,
    TimeMapAxis,
    WcsGeom,
    WcsNDMap,
)


######################################################################
# Creating WCS Maps
# -----------------
#
# Using Factory Methods
# ~~~~~~~~~~~~~~~~~~~~~
#
# Maps are most easily created using the `~gammapy.maps.Map.create`
# factory method:
#

m_allsky = Map.create()


######################################################################
# Calling `~gammapy.maps.Map.create` without any further arguments creates by
# default an allsky WCS map using a CAR projection, ICRS coordinates and a
# pixel size of 1 deg. This can be easily checked by printing the
# `~gammapy.maps.Map.geom` attribute of the map:
#

print(m_allsky.geom)


######################################################################
# The `~gammapy.maps.Map.geom` attribute is a `~gammapy.maps.Geom` object, that defines the basic
# geometry of the map, such as size of the pixels, width and height of the
# image, coordinate system etc., but we will learn more about this object
# later.
#
# Besides the ``.geom`` attribute the map has also a ``.data`` attribute,
# which is just a plain `numpy.ndarray` and stores the data associated
# with this map:
#

print(m_allsky.data)


######################################################################
# By default maps are filled with zeros.
#
# The ``map_type`` argument can be used to control the pixelization scheme
# (WCS or HPX).
#

position = SkyCoord(0.0, 5.0, frame="galactic", unit="deg")

# Create a WCS Map
m_wcs = Map.create(binsz=0.1, map_type="wcs", skydir=position, width=10.0)

# Create a HPX Map
m_hpx = Map.create(binsz=0.1, map_type="hpx", skydir=position, width=10.0)


######################################################################
# Here is an example that creates a WCS map centered on the Galactic
# center and now uses Galactic coordinates:
#

skydir = SkyCoord(0, 0, frame="galactic", unit="deg")
m_gc = Map.create(
    binsz=0.02, width=(10, 5), skydir=skydir, frame="galactic", proj="TAN"
)
print(m_gc.geom)


######################################################################
# In addition we have defined a TAN projection, a pixel size of ``0.02``
# deg and a width of the map of ``10 deg x 5 deg``. The ``width`` argument
# also takes scalar value instead of a tuple, which is interpreted as both
# the width and height of the map, so that a quadratic map is created.
#


######################################################################
# Creating from a Map Geometry
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# As we have seen in the first examples, the `~gammapy.maps.Map` object couples the
# data (stored as a `~numpy.ndarray`) with a `~gammapy.maps.Geom` object. The
# `~gammapy.maps.Geom` object can be seen as a generalization of an
# `astropy.wcs.WCS` object, providing the information on how the data
# maps to physical coordinate systems. In some cases e.g. when creating
# many maps with the same WCS geometry it can be advantageous to first
# create the map geometry independent of the map object it-self:
#

wcs_geom = WcsGeom.create(binsz=0.02, width=(10, 5), skydir=(0, 0), frame="galactic")


######################################################################
# And then create the map objects from the ``wcs_geom`` geometry
# specification:
#

maps = {}

for name in ["counts", "background"]:
    maps[name] = Map.from_geom(wcs_geom)


######################################################################
# The `~gammapy.maps.Geom` object also has a few helpful methods. E.g. we can check
# whether a given position on the sky is contained in the map geometry:
#

# define the position of the Galactic center and anti-center
positions = SkyCoord([0, 180], [0, 0], frame="galactic", unit="deg")
wcs_geom.contains(positions)


######################################################################
# Or get the image center of the map:
#

print(wcs_geom.center_skydir)


######################################################################
# Or we can also retrieve the solid angle per pixel of the map:
#

print(wcs_geom.solid_angle())


######################################################################
# Adding Non-Spatial Axes
# ~~~~~~~~~~~~~~~~~~~~~~~
#
# In many analysis scenarios we would like to add extra dimension to the
# maps to study e.g. energy or time dependency of the data. Those
# non-spatial dimensions are handled with the `~gammapy.maps.MapAxis` object. Let us
# first define an energy axis, with 4 bins:
#

energy_axis = MapAxis.from_bounds(
    1, 100, nbin=4, unit="TeV", name="energy", interp="log"
)
print(energy_axis)


######################################################################
# Where ``interp='log'`` specifies that a logarithmic spacing is used
# between the bins, equivalent to ``np.logspace(0, 2, 4)``. This
# `~gammapy.maps.MapAxis` object we can now pass to `~gammapy.maps.Map.create()` using the
# ``axes=`` argument:
#

m_cube = Map.create(binsz=0.02, width=(10, 5), frame="galactic", axes=[energy_axis])
print(m_cube.geom)


######################################################################
# Now we see that besides ``lon`` and ``lat`` the map has an additional
# axes named ``energy`` with 4 bins. The total dimension of the map is now
# ``ndim=3``.
#
# We can also add further axes by passing a list of `~gammapy.maps.MapAxis` objects.
# To demonstrate this we create a time axis with linearly spaced bins and
# pass both axes to `~gammapy.maps.Map.create()`:
#

time_axis = MapAxis.from_bounds(0, 24, nbin=24, unit="hour", name="time", interp="lin")

m_4d = Map.create(
    binsz=0.02, width=(10, 5), frame="galactic", axes=[energy_axis, time_axis]
)
print(m_4d.geom)


######################################################################
# The `~gammapy.maps.MapAxis` object internally stores the coordinates or “position
# values” associated with every map axis bin or “node”. We distinguish
# between two node types: ``"edges"`` and ``"center"``. The node type
# ``"edges"`` (which is also the default) specifies that the data
# associated with this axis is integrated between the edges of the bin
# (e.g. counts data). The node type ``"center"`` specifies that the data is
# given at the center of the bin (e.g. exposure or differential fluxes).
#
# The edges of the bins can be checked with `~gammapy.maps.MapAxis.edges` attribute:
#

print(energy_axis.edges)


######################################################################
# The numbers are given in the units we specified above, which can be
# checked again with:
#

print(energy_axis.unit)


######################################################################
# The centers of the axis bins can be checked with the `~gammapy.maps.MapAxis.center`
# attribute:
#

print(energy_axis.center)

######################################################################
# Adding Non-contiguous axes
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Non-spatial map axes can also be handled through two other objects known as the `~gammapy.maps.TimeMapAxis`
# and the `~gammapy.maps.LabelMapAxis`.
#


######################################################################
# TimeMapAxis
# ^^^^^^^^^^^
#
# The `~gammapy.maps.TimeMapAxis` object provides an axis for non-adjacent
# time intervals.
#

time_map_axis = TimeMapAxis(
    edges_min=[1, 5, 10, 15] * u.day,
    edges_max=[2, 7, 13, 18] * u.day,
    reference_time=Time("2020-03-19"),
)

print(time_map_axis)

######################################################################
# This ``time_map_axis`` can then be utilised in a similar way to the previous implementation to create
# a `~gammapy.maps.Map`.
#

map_4d = Map.create(
    binsz=0.02, width=(10, 5), frame="galactic", axes=[energy_axis, time_map_axis]
)
print(map_4d.geom)

######################################################################
# It is possible to utilise the `~gammapy.maps.TimeMapAxis.slice` attrribute
# to create new a `~gammapy.maps.TimeMapAxis`. Here we are slicing
# between the first and third axis to extract the subsection of the axis
# between indice 0 and 2.

print(time_map_axis.slice([0, 2]))

######################################################################
# It is also possible to `~gammapy.maps.TimeMapAxis.squash` the axis,
# which squashes the existing axis into one bin. This creates a new axis
# between the extreme edges of the initial axis.

print(time_map_axis.squash())


######################################################################
# The `~gammapy.maps.TimeMapAxis.is_contiguous` method returns a boolean
# which indicates whether the `~gammapy.maps.TimeMapAxis` is contiguous or not.

print(time_map_axis.is_contiguous)

######################################################################
# As we have a non-contiguous axis we can print the array of bin edges for both
# the minimum axis edges (`~gammapy.maps.TimeMapAxis.edges_min`) and the maximum axis
# edges (`~gammapy.maps.TimeMapAxis.edges_max`).

print(time_map_axis.edges_min)

print(time_map_axis.edges_max)

######################################################################
# Next, we use the `~gammapy.maps.TimeMapAxis.to_contiguous` functionality to
# create a contiguous axis and expose `~gammapy.maps.TimeMapAxis.edges`. This
# method returns a `~astropy.units.Quantity` with respect to the reference time.

time_map_axis_contiguous = time_map_axis.to_contiguous()

print(time_map_axis_contiguous.is_contiguous)

print(time_map_axis_contiguous.edges)


######################################################################
# The `~gammapy.maps.TimeMapAxis.time_edges` will return the `~astropy.time.Time` object directly

print(time_map_axis_contiguous.time_edges)


######################################################################
# `~gammapy.maps.TimeMapAxis` also has both functionalities of
# `~gammapy.maps.TimeMapAxis.coord_to_pix` and `~gammapy.maps.TimeMapAxis.coord_to_idx`.
# The `~gammapy.maps.TimeMapAxis.coord_to_idx` attribute will give the index of the
# ``time`` specified, similarly for `~gammapy.maps.TimeMapAxis.coord_to_pix` which returns
# the pixel. A linear interpolation is assumed.
#
# Start by choosing a time which we know is within the `~gammapy.maps.TimeMapAxis` and see the results.


time = Time(time_map_axis.time_max.mjd[0], format="mjd")

print(time_map_axis.coord_to_pix(time))

print(time_map_axis.coord_to_idx(time))


######################################################################
# This functionality can also be used with an array of `~astropy.time.Time` values.

times = Time(time_map_axis.time_max.mjd, format="mjd")

print(time_map_axis.coord_to_pix(times))

print(time_map_axis.coord_to_idx(times))

######################################################################
# Note here we take a `~astropy.time.Time` which is outside the edges.
# A linear interpolation is assumed for both methods, therefore for a time
# outside the ``time_map_axis`` there is no extrapolation and -1 is returned.
#
# Note: due to this, the `~gammapy.maps.TimeMapAxis.coord_to_pix` method will
# return ``nan`` and the `~gammapy.maps.TimeMapAxis.coord_to_idx` method returns -1.

print(time_map_axis.coord_to_pix(Time(time.mjd + 1, format="mjd")))

print(time_map_axis.coord_to_idx(Time(time.mjd + 1, format="mjd")))


######################################################################
# LabelMapAxis
# ^^^^^^^^^^^^
#
# The `~gammapy.maps.LabelMapAxis` object allows for handling of labels for map axes.
# It provides an axis for non-numeric entries.
#

label_axis = LabelMapAxis(
    labels=["dataset-1", "dataset-2", "dataset-3"], name="dataset"
)

print(label_axis)

######################################################################
# The labels can be checked using the `~gammapy.maps.LabelMapAxis.center` attribute:
print(label_axis.center)


######################################################################
# To obtain the position of the label, one can utilise the `~gammapy.maps.LabelMapAxis.coord_to_pix` attribute

print(label_axis.coord_to_pix(["dataset-3"]))

######################################################################
# To adapt and create new axes the following attributes can be utilised:
# `~gammapy.maps.LabelMapAxis.concatenate`, `~gammapy.maps.LabelMapAxis.slice` and
# `~gammapy.maps.LabelMapAxis.squash`.
#
# Combining two different `~gammapy.maps.LabelMapAxis` is done in the following way:

label_axis2 = LabelMapAxis(labels=["dataset-a", "dataset-b"], name="dataset")

print(label_axis.concatenate(label_axis2))

######################################################################
# A new `~gammapy.maps.LabelMapAxis` can be created by slicing an already existing one.
# Here we are slicing between the second and third bins to extract the subsection.
print(label_axis.slice([1, 2]))

######################################################################
# A new axis object can be created by squashing the axis into a single bin.

print(label_axis.squash())


######################################################################
# Mixing the three previous axis types (`~gammapy.maps.MapAxis`,
# `~gammapy.maps.TimeMapAxis` and `~gammapy.maps.LabelMapAxis`)
# would be done like so:
#

axes = MapAxes(axes=[energy_axis, time_map_axis, label_axis])
hdu = axes.to_table_hdu(format="gadf")
table = Table.read(hdu)
display(table)


######################################################################
# Reading and Writing
# -------------------
#
# Gammapy `~gammapy.maps.Map` objects are serialized using the Flexible Image
# Transport Format (FITS). Depending on the pixelization scheme (HEALPix
# or WCS) and presence of non-spatial dimensions the actual convention to
# write the FITS file is different. By default Gammapy uses a generic
# convention named ``"gadf"``, which will support WCS and HEALPix formats as
# well as an arbitrary number of non-spatial axes. The convention is
# documented in detail on the `Gamma Astro Data
# Formats <https://gamma-astro-data-formats.readthedocs.io/en/latest/skymaps/index.html>`__
# page.
#
# Other conventions required by specific software (e.g. the Fermi Science
# Tools) are supported as well. At the moment those are the following
#
# -  ``"fgst-ccube"``: Fermi counts cube format.
# -  ``"fgst-ltcube"``: Fermi livetime cube format.
# -  ``"fgst-bexpcube"``: Fermi exposure cube format
# -  ``"fgst-template"``: Fermi Galactic diffuse and source template format.
# -  ``"fgst-srcmap"`` and ``"fgst-srcmap-sparse"``: Fermi source map and
#    sparse source map format.
#
# The conventions listed above only support an additional energy axis.
#
# Reading Maps
# ~~~~~~~~~~~~
#
# Reading FITS files is mainly exposed via the `~gammapy.maps.Map.read()` method. Let
# us take a look at a first example:
#

filename = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-counts.fits.gz"
m_3fhl_gc = Map.read(filename)
print(m_3fhl_gc)


######################################################################
# If ``map_type`` argument is not given when calling read a map object
# will be instantiated with the pixelization of the input HDU.
#
# By default ``Map.read()`` will try to find the first valid data hdu in
# the filename and read the data from there. If multiple HDUs are present
# in the FITS file, the desired one can be chosen with the additional
# `hdu=` argument:
#

m_3fhl_gc = Map.read(filename, hdu="PRIMARY")
print(m_3fhl_gc)


######################################################################
# In rare cases e.g. when the FITS file is not valid or meta data is
# missing from the header it can be necessary to modify the header of a
# certain HDU before creating the `~gammapy.maps.Map` object. In this case we can use
# `astropy.io.fits` directly to read the FITS file:
#

filename = os.environ["GAMMAPY_DATA"] + "/fermi-3fhl-gc/fermi-3fhl-gc-exposure.fits.gz"
hdulist = fits.open(filename)
print(hdulist.info())


######################################################################
# And then modify the header keyword and use `~gammapy.maps.from_hdulist()` to
# create the `~gammapy.maps.Map` object after:
#

hdulist["PRIMARY"].header["BUNIT"] = "cm2 s"
print(Map.from_hdulist(hdulist=hdulist))


######################################################################
# Writing Maps
# ~~~~~~~~~~~~
#
# Writing FITS files on disk via the `~gammapy.maps.Map.write()` method.
# Here is a first example:
#

m_cube.write("example_cube.fits", overwrite=True)


######################################################################
# By default Gammapy does not overwrite files. In this example we set
# ``overwrite=True`` in case the cell gets executed multiple times. Now we
# can read back the cube from disk using `~gammapy.maps.Map.read()`:
#

m_cube = Map.read("example_cube.fits")
print(m_cube)


######################################################################
# We can also choose a different FITS convention to write the example cube
# in a format compatible to the Fermi Galactic diffuse background model:
#

m_cube.write("example_cube_fgst.fits", format="fgst-template", overwrite=True)


######################################################################
# To understand a little bit better the generic ``gadf`` convention we use
# `~gammapy.maps.Map.to_hdulist()` to generate a list of FITS HDUs first:
#

hdulist = m_4d.to_hdulist(format="gadf")
print(hdulist.info())


######################################################################
# As we can see the ``HDUList`` object contains to HDUs. The first one
# named ``PRIMARY`` contains the data array with shape corresponding to
# our data and the WCS information stored in the header:
#

print(hdulist["PRIMARY"].header)


######################################################################
# The second HDU is a ``BinTableHDU`` named ``PRIMARY_BANDS`` contains the
# information on the non-spatial axes such as name, order, unit, min, max
# and center values of the axis bins. We use an `astropy.table.Table` to
# show the information:
#

print(Table.read(hdulist["PRIMARY_BANDS"]))


######################################################################
# Maps can be serialized to a sparse data format by calling write with
# ``sparse=True``. This will write all non-zero pixels in the map to a
# data table appropriate to the pixelization scheme.
#

m = Map.create(binsz=0.1, map_type="wcs", width=10.0)
m.write("file.fits", hdu="IMAGE", sparse=True, overwrite=True)
m = Map.read("file.fits", hdu="IMAGE", map_type="wcs")


######################################################################
# Accessing Data
# --------------
#
# How to get data values
# ~~~~~~~~~~~~~~~~~~~~~~
#
# All map objects have a set of accessor methods, which can be used to
# access or update the contents of the map irrespective of its underlying
# representation. Those accessor methods accept as their first argument a
# coordinate `tuple` containing scalars, `list`, or `numpy.ndarray`
# with one tuple element for each dimension. Some methods additionally
# accept a `dict` or `~gammapy.maps.MapCoord` argument, of which both allow to
# assign coordinates by axis name.
#
# Let us first begin with the `~gammapy.maps.Map.get_by_idx()` method, that accepts a
# tuple of indices. The order of the indices corresponds to the axis order
# of the map:
#

print(m_gc.get_by_idx((50, 30)))


######################################################################
# **Important:** Gammapy uses a reversed index order in the map API with
# the longitude axes first. To achieve the same by directly indexing into
# the numpy array we have to call:
#

print(m_gc.data[([30], [50])])


######################################################################
# To check the order of the axes you can always print the ``.geom``
# attribute:
#

print(m_gc.geom)


######################################################################
# To access values directly by sky coordinates we can use the
# `~gammapy.maps.Map.get_by_coord()` method. This time we pass in a `dict`, specifying
# the axes names corresponding to the given coordinates:
#

print(m_gc.get_by_coord({"lon": [0, 180], "lat": [0, 0]}))


######################################################################
# The units of the coordinates are assumed to be in degrees in the
# coordinate system used by the map. If the coordinates do not correspond
# to the exact pixel center, the value of the nearest pixel center will be
# returned. For positions outside the map geometry `numpy.nan` is returned.
#
# The coordinate or idx arrays follow normal `Numpy broadcasting
# rules <https://jakevdp.github.io/PythonDataScienceHandbook/02.05-computation-on-arrays-broadcasting.html>`__.
# So the following works as expected:
#

lons = np.linspace(-4, 4, 10)
print(m_gc.get_by_coord({"lon": lons, "lat": 0}))


######################################################################
# Or as an even more advanced example, we can provide ``lats`` as column
# vector and broadcasting to a 2D result array will be applied:
#

lons = np.linspace(-4, 4, 8)
lats = np.linspace(-4, 4, 8).reshape(-1, 1)
print(m_gc.get_by_coord({"lon": lons, "lat": lats}))


######################################################################
# Indexing and Slicing Sub-Maps
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# When you have worked with Numpy arrays in the past you are probably
# familiar with the concept of indexing and slicing into data arrays. To
# support slicing of non-spatial axes of `~gammapy.maps.Map` objects, the `~gammapy.maps.Map`
# object has a `~gammapy.maps.Map.slice_by_idx()` method, which allows to extract
# sub-maps from a larger map.
#
# The following example demonstrates how to get the map at the energy bin
# number 3:
#

m_sub = m_cube.slice_by_idx({"energy": 3})
print(m_sub)


######################################################################
# Note that the returned object is again a `~gammapy.maps.Map` with updated axes
# information. In this case, because we extracted only a single image, the
# energy axes is dropped from the map.
#
# To extract a sub-cube with a sliced energy axes we can use a normal
# ``slice`` object:
#

m_sub = m_cube.slice_by_idx({"energy": slice(1, 3)})
print(m_sub)


######################################################################
# Note that the returned object is also a `~gammapy.maps.Map` object, but this time
# with updated energy axis specification.
#
# Slicing of multiple dimensions is supported by adding further entries to
# the dict passed to `~gammapy.maps.Map.slice_by_idx()`
#

m_sub = m_4d.slice_by_idx({"energy": slice(1, 3), "time": slice(4, 10)})
print(m_sub)


######################################################################
# For convenience there is also a `~gammapy.maps.Map.get_image_by_coord()` method which
# allows to access image planes at given non-spatial physical coordinates.
# This method also supports `~astropy.units.Quantity` objects:
#

image = m_4d.get_image_by_coord({"energy": 4 * u.TeV, "time": 5 * u.h})
print(image.geom)


######################################################################
# Iterating by image
# ~~~~~~~~~~~~~~~~~~
#
# For maps with non-spatial dimensions the `~gammapy.maps.Map.iter_by_image_data`
# method can be used to loop over image slices. The image plane index
# ``idx`` is returned in data order, so that the data array can be indexed
# directly. Here is an example for an in-place convolution of an image
# using `~astropy.convolution.convolve` to interpolate NaN values:
#

axis1 = MapAxis([1, 10, 100], interp="log", name="energy")
axis2 = MapAxis([1, 2, 3], interp="lin", name="time")
m = Map.create(width=(5, 3), axes=[axis1, axis2], binsz=0.1)
m.data[:, :, 15:18, 20:25] = np.nan

for img, idx in m.iter_by_image_data():
    kernel = np.ones((5, 5))
    m.data[idx] = convolve(img, kernel)

assert not np.isnan(m.data).any()


######################################################################
# Modifying Data
# --------------
#
# How to set data values
# ~~~~~~~~~~~~~~~~~~~~~~
#
# To modify and set map data values the `~gammapy.maps.Map` object features as well a
# `~gammapy.maps.Map.set_by_idx()` method:
#

m_cube.set_by_idx(idx=(10, 20, 3), vals=42)


######################################################################
# here we check that data have been updated:
#

print(m_cube.get_by_idx((10, 20, 3)))


######################################################################
# Of course there is also a `~gammapy.maps.Map.set_by_coord()` method, which allows to
# set map data values in physical coordinates.
#

m_cube.set_by_coord({"lon": 0, "lat": 0, "energy": 2 * u.TeV}, vals=42)


######################################################################
# Again the ``lon`` and ``lat`` values are assumed to be given in degrees
# in the coordinate system used by the map. For the energy axis, the unit
# is the one specified on the axis (use ``m_cube.geom.axes[0].unit`` to
# check if needed…).
#
# All ``.xxx_by_coord()`` methods accept `~astropy.coordinates.SkyCoord` objects as well. In
# this case we have to use the ``"skycoord"`` keyword instead of ``"lon"`` and
# ``"lat"``:
#

skycoords = SkyCoord([1.2, 3.4], [-0.5, 1.1], frame="galactic", unit="deg")
m_cube.set_by_coord({"skycoord": skycoords, "energy": 2 * u.TeV}, vals=42)


######################################################################
# Filling maps from event lists
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# This example shows how to fill a counts cube from an event list:
#

energy_axis = MapAxis.from_bounds(
    10.0, 2e3, 12, interp="log", name="energy", unit="GeV"
)
counts_3d = WcsNDMap.create(
    binsz=0.1, width=10.0, skydir=(0, 0), frame="galactic", axes=[energy_axis]
)

events = EventList.read("$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz")

counts_3d.fill_by_coord({"skycoord": events.radec, "energy": events.energy})
counts_3d.write("ccube.fits", format="fgst-ccube", overwrite=True)


######################################################################
# Alternatively you can use the `~gammapy.maps.Map.fill_events` method:
#

counts_3d = WcsNDMap.create(
    binsz=0.1, width=10.0, skydir=(0, 0), frame="galactic", axes=[energy_axis]
)

counts_3d.fill_events(events)


######################################################################
# If you have a given map already, and want to make a counts image with
# the same geometry (not using the pixel data from the original map), you
# can also use the `~gammapy.maps.Map.fill_events` method.
#

events = EventList.read("$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz")
reference_map = Map.read("$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-counts.fits.gz")
counts = Map.from_geom(reference_map.geom)
counts.fill_events(events)


######################################################################
# It works for IACT and Fermi-LAT events, for WCS or HEALPix map
# geometries, and also for extra axes. Especially energy axes are
# automatically handled correctly.
#


######################################################################
# Filling maps from interpolation
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Maps support interpolation via the `~~gammapy.maps.Map.interp_by_coord` and
# `~~gammapy.maps.Map.interp_by_pix` methods. Currently, the following interpolation
# methods are supported:
#
# -  ``"nearest"`` : Return value of nearest pixel (no interpolation).
# -  ``"linear"`` : Interpolation with first order polynomial. This is the
#    only interpolation method that is supported for all map types.
# -  ``quadratic`` : Interpolation with second order polynomial.
# -  ``cubic`` : Interpolation with third order polynomial.
#
# Note that ``"quadratic"`` and ``"cubic"`` interpolation are currently only
# supported for WCS-based maps with regular geometry (e.g. 2D or ND with
# the same geometry in every image plane). ``"linear"`` and higher order
# interpolation by pixel coordinates is only supported for WCS-based maps.
#
# In the following example we create a new map and fill it by
# interpolating another map:
#

# read map
filename = "$GAMMAPY_DATA/fermi-3fhl-gc/gll_iem_v06_gc.fits.gz"
m_iem_gc = Map.read(filename)

# create new geometry
skydir = SkyCoord(266.4, -28.9, frame="icrs", unit="deg")
wcs_geom_cel = WcsGeom.create(skydir=skydir, binsz=0.1, frame="icrs", width=(8, 4))

# create new empty map from geometry
m_iem_10GeV = Map.from_geom(wcs_geom_cel)
coords = m_iem_10GeV.geom.get_coord()

# fill new map using interpolation
m_iem_10GeV.data = m_iem_gc.interp_by_coord(
    {"skycoord": coords.skycoord, "energy_true": 10 * u.GeV},
    method="linear",
    fill_value=np.nan,
)


######################################################################
# Interpolating onto a different geometry
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#


######################################################################
# For 3d geometries this operation can be performed directly using the
# `~gammapy.maps.Map.interp_to_geom()` method. This is very useful, ex: while using map
# arithmetic.
#

# create new geometry
energy_axis = MapAxis.from_bounds(
    10.0, 2e3, 6, interp="log", name="energy_true", unit="GeV"
)
skydir = SkyCoord(266.4, -28.9, frame="icrs", unit="deg")
wcs_geom_3d = WcsGeom.create(
    skydir=skydir, binsz=0.1, frame="icrs", width=(8, 4), axes=[energy_axis]
)

# create the interpolated map
m_iem_interp = m_iem_gc.interp_to_geom(
    wcs_geom_3d, preserve_counts=False, method="linear", fill_value=np.nan
)
print(m_iem_interp)


######################################################################
# Note that ``preserve_counts=`` option should be true if the map is an
# integral quantity (e.g. counts) and false if the map is a differential
# quantity (e.g. intensity).
#


######################################################################
# Maps operations
# ---------------
#
# Basic operators
# ~~~~~~~~~~~~~~~
#
# One can perform simple arithmetic on maps using the ``+``, ``-``, ``*``,
# ``/`` operators, this works only for maps with the same geometry:
#

iem_plus_iem = m_iem_10GeV + m_iem_10GeV

iem_minus_iem = m_iem_10GeV - m_iem_10GeV


######################################################################
# These operations can be applied between a Map and a scalar in that
# specific order:
#

iem_times_two = m_iem_10GeV * 2
# iem_times_two = 2 * m_iem_10GeV # this won't work


######################################################################
# The logic operators can also be applied on maps (the result is a map of
# boolean type):
#

is_null = iem_minus_iem == 0
print(is_null)


######################################################################
# Here we check that the result is ``True`` for all the well-defined
# pixels (not ``NaN``):
#

print(np.all(is_null.data[~np.isnan(iem_minus_iem)]))


######################################################################
# Cutouts
# ~~~~~~~
#
# The `~gammapy.maps.WcsNDMap` objects features a `~gammapy.maps.WcsNDMap.cutout()` method, which allows
# you to cut out a smaller part of a larger map. This can be useful,
# e.g. when working with all-sky diffuse maps. Here is an example:
#

position = SkyCoord(0, 0, frame="galactic", unit="deg")
m_iem_cutout = m_iem_gc.cutout(position=position, width=(4 * u.deg, 2 * u.deg))


######################################################################
# The returned object is again a `~gammapy.maps.Map` object with updated WCS
# information and data size. As one can see the cutout is automatically
# applied to all the non-spatial axes as well. The cutout width is given
# in the order of ``(lon, lat)`` and can be specified with units that will
# be handled correctly.
#


######################################################################
# Visualizing and Plotting
# ------------------------
#
# All map objects provide a `~gammapy.maps.Map.plot()` method for generating a visualization
# of a map. This method returns figure, axes, and image objects that can
# be used to further tweak/customize the image. The `~gammapy.maps.Map.plot()` method should
# be used with 2D maps, while 3D maps can be displayed with the
# `~gammapy.maps.Map.plot_interactive()` or `~gammapy.maps.Map.plot_grid()` methods.
#
# Image Plotting
# ~~~~~~~~~~~~~~
#
# For debugging and inspecting the map data it is useful to plot or
# visualize the images planes contained in the map.
#

filename = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-counts.fits.gz"
m_3fhl_gc = Map.read(filename)


######################################################################
# After reading the map we can now plot it on the screen by calling the
# ``.plot()`` method:
#

m_3fhl_gc.plot()
plt.show()


######################################################################
# We can easily improve the plot by calling `~gammapy.maps.Map.smooth()` first and
# providing additional arguments to `~gammapy.maps.Map.plot()`. Most of them are passed
# further to
# `plt.imshow() <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.imshow.html>`__:
#

smoothed = m_3fhl_gc.smooth(width=0.2 * u.deg, kernel="gauss")
smoothed.plot(stretch="sqrt", add_cbar=True, vmax=4, cmap="inferno")
plt.show()


######################################################################
# We can use the
# `plt.rc_context() <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.rc_context.html>`__
# context manager to further tweak the plot by adapting the figure and
# font size:
#

rc_params = {"figure.figsize": (12, 5.4), "font.size": 12}
with plt.rc_context(rc=rc_params):
    smoothed = m_3fhl_gc.smooth(width=0.2 * u.deg, kernel="gauss")
    smoothed.plot(stretch="sqrt", add_cbar=True, vmax=4)
plt.show()


######################################################################
# Cube plotting
# ~~~~~~~~~~~~~
#
# For maps with non-spatial dimensions the `~gammapy.maps.Map` object features an
# interactive plotting method, that works in jupyter notebooks only (Note:
# it requires the package ``ipywidgets`` to be installed). We first read a
# small example cutout from the Fermi Galactic diffuse model and display
# the data cube by calling `~gammapy.maps.Map.plot_interactive()`:
#

rc_params = {
    "figure.figsize": (12, 5.4),
    "font.size": 12,
    "axes.formatter.limits": (2, -2),
}
m_iem_gc.plot_interactive(add_cbar=True, stretch="sqrt", rc_params=rc_params)
plt.show()


######################################################################
# Now you can use the interactive slider to select an energy range and the
# corresponding image is displayed on the screen. You can also use the
# radio buttons to select your preferred image stretching. We have passed
# additional keywords using the ``rc_params`` argument to improve the
# figure and font size. Those keywords are directly passed to the
# `plt.rc_context() <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.rc_context.html>`__
# context manager.
#
# Additionally, all the slices of a 3D `~gammapy.maps.Map` can be displayed using the
# `~gammapy.maps.Map.plot_grid()` method. By default the colorbars bounds of the subplots
# are not the same, we can make them consistent using the ``vmin`` and
# ``vmax`` options:
#

counts_3d.plot_grid(ncols=4, figsize=(16, 12), vmin=0, vmax=100, stretch="log")
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html

__all__ = ["Registry"]


class Registry(list):
    """Registry class."""

    def get_cls(self, tag):
        for cls in self:
            tags = getattr(cls, "tag", [])
            tags = [tags] if isinstance(tags, str) else tags
            if tag in tags:
                return cls
        raise KeyError(f"No object found with tag: {tag!r}")

    def __str__(self):
        info = "Registry\n"
        info += "--------\n\n"
        len_max = max([len(_.__name__) for _ in self])

        for item in self:
            info += f"{item.__name__:{len_max}s}: {item.tag} \n"

        return info.expandtabs(tabsize=2)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"



"""Plot Fermi PSF."""

from gammapy.irf import PSFMap
from gammapy.maps import MapAxis, WcsGeom

filename = "$GAMMAPY_DATA/fermi_3fhl/fermi_3fhl_psf_gc.fits.gz"
psf = PSFMap.read(filename, format="gtpsf")

axis = MapAxis.from_energy_bounds("10 GeV", "2 TeV", nbin=20, name="energy_true")
geom = WcsGeom.create(npix=50, binsz=0.01, axes=[axis])

# .to_image() computes the exposure weighted mean PSF
kernel = psf.get_psf_kernel(geom=geom).to_image()

kernel.psf_kernel_map.plot()



r"""
.. _powerlaw-spectral-model:

Power law spectral model
========================

This model parametrises a power law spectrum.

It is defined by the following equation:

.. math::
    \phi(E) = \phi_0 \cdot \left( \frac{E}{E_0} \right)^{-\Gamma}

"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.modeling.models import Models, PowerLawSpectralModel, SkyModel

energy_bounds = [0.1, 100] * u.TeV
model = PowerLawSpectralModel(
    index=2,
    amplitude="1e-12 TeV-1 cm-2 s-1",
    reference=1 * u.TeV,
)
model.plot(energy_bounds)
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(spectral_model=model, name="power-law-model")
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Estimators."""

from gammapy.utils.registry import Registry
from .core import Estimator
from .energydependentmorphology import EnergyDependentMorphologyEstimator
from .map import ASmoothMapEstimator, ExcessMapEstimator, FluxMaps, TSMapEstimator
from .metadata import FluxMetaData
from .parameter import ParameterEstimator, ParameterSensitivityEstimator
from .points import (
    FluxPoints,
    FluxPointsEstimator,
    FluxProfileEstimator,
    LightCurveEstimator,
    SensitivityEstimator,
)
from .profile import ImageProfile, ImageProfileEstimator

__all__ = [
    "ASmoothMapEstimator",
    "Estimator",
    "ESTIMATOR_REGISTRY",
    "ExcessMapEstimator",
    "FluxMaps",
    "FluxPoints",
    "FluxPointsEstimator",
    "FluxProfileEstimator",
    "ImageProfile",
    "ImageProfileEstimator",
    "LightCurveEstimator",
    "ParameterEstimator",
    "SensitivityEstimator",
    "ParameterSensitivityEstimator",
    "TSMapEstimator",
    "EnergyDependentMorphologyEstimator",
    "FluxMetaData",
]


ESTIMATOR_REGISTRY = Registry(
    [
        ExcessMapEstimator,
        TSMapEstimator,
        FluxPointsEstimator,
        ASmoothMapEstimator,
        LightCurveEstimator,
        SensitivityEstimator,
        ParameterSensitivityEstimator,
        FluxProfileEstimator,
        ParameterEstimator,
    ]
)
"""Registry of estimator classes in Gammapy."""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Tools to create profiles (i.e. 1D "slices" from 2D images)."""

import logging
from itertools import repeat
import numpy as np
from astropy import units as u
from regions import CircleAnnulusSkyRegion
import gammapy.utils.parallel as parallel
from gammapy.datasets import Datasets
from gammapy.maps import MapAxis
from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
from .core import FluxPoints
from .sed import FluxPointsEstimator

log = logging.getLogger(__name__)

__all__ = ["FluxProfileEstimator"]


class FluxProfileEstimator(FluxPointsEstimator):
    """Estimate flux profiles.

    The class is backward folding of FluxPointsEstimator. However, the re-optimization
    is not available, as only one spectral model can be fitted.

    Parameters
    ----------
    regions : list of `~regions.SkyRegion`
        Regions to use.
    spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
        Spectral model to compute the fluxes or brightness.
        Default is power-law with spectral index of 2.
    n_jobs : int, optional
        Number of processes used in parallel for the computation. Default is one,
        unless `~gammapy.utils.parallel.N_JOBS_DEFAULT` was modified. The number
        of jobs is limited to the number of physical CPUs.
    parallel_backend : {"multiprocessing", "ray"}, optional
        Which backend to use for multiprocessing.
        Defaults to `~gammapy.utils.parallel.BACKEND_DEFAULT`.
    **kwargs : dict, optional
        Keywords forwarded to the `FluxPointsEstimator` (see documentation
        there for further description of valid keywords). Note that the `reoptimized`
        keyword is accepted only if set to False. If not, an error is raised.
        If the keyword is not set by the user, it will be internally set to False
        by default.

    Examples
    --------
    This example shows how to compute a counts profile for the Fermi galactic
    center region::

    >>> from astropy import units as u
    >>> from astropy.coordinates import SkyCoord
    >>> from gammapy.data import GTI
    >>> from gammapy.estimators import FluxProfileEstimator
    >>> from gammapy.utils.regions import make_orthogonal_rectangle_sky_regions
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.maps import RegionGeom

    >>> # load example data
    >>> filename = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc.fits.gz"
    >>> dataset = MapDataset.read(filename, name="fermi-dataset")

    >>> # configuration
    >>> dataset.gti = GTI.create("0s", "1e7s", "2010-01-01")

    >>> # creation of the boxes and axis
    >>> start_pos = SkyCoord("-1d", "0d", frame='galactic')
    >>> end_pos = SkyCoord("1d", "0d", frame='galactic')

    >>> regions = make_orthogonal_rectangle_sky_regions(
    ...            start_pos=start_pos,
    ...            end_pos=end_pos,
    ...            wcs=dataset.counts.geom.wcs,
    ...            height=2 * u.deg,
    ...            nbin=21
    ...        )

    >>> # set up profile estimator and run
    >>> prof_maker = FluxProfileEstimator(regions=regions, energy_edges=[10, 2000] * u.GeV)
    >>> fermi_prof = prof_maker.run(dataset)
    >>> print(fermi_prof)
    FluxPoints
    ----------
    <BLANKLINE>
      geom                   : RegionGeom
      axes                   : ['lon', 'lat', 'energy', 'projected-distance']
      shape                  : (1, 1, 1, 21)
      quantities             : ['norm', 'norm_err', 'ts', 'npred', 'npred_excess', 'stat', 'stat_null', 'counts', 'success']
      ref. model             : pl
      n_sigma                : 1
      n_sigma_ul             : 2
      sqrt_ts_threshold_ul   : 2
      sed type init          : likelihood

    """

    tag = "FluxProfileEstimator"

    def __init__(self, regions, spectral_model=None, **kwargs):
        if len(regions) <= 1:
            raise ValueError(
                "Please provide at least two regions for flux profile estimation."
            )

        self.regions = regions

        if spectral_model is None:
            spectral_model = PowerLawSpectralModel()

        self.spectral_model = spectral_model

        reoptimize = kwargs.get("reoptimize", False)
        if reoptimize is True:
            raise ValueError(
                f"reoptimize=True is not available for {self}. Set to `False` instead."
            )
        super().__init__(**kwargs)

    @property
    def projected_distance_axis(self):
        """Get projected distance from the first region.

        For normal region this is defined as the distance from the
        center of the region. For annulus shaped regions it is the
        mean between the inner and outer radius.

        Returns
        -------
        axis : `MapAxis`
            Projected distance axis.
        """
        distances = []
        center = self.regions[0].center

        for idx, region in enumerate(self.regions):
            if isinstance(region, CircleAnnulusSkyRegion):
                distance = (region.inner_radius + region.outer_radius) / 2.0
            else:
                distance = center.separation(region.center)

            distances.append(distance)

        return MapAxis.from_nodes(
            u.Quantity(distances, "deg"), name="projected-distance"
        )

    def run(self, datasets):
        """Run flux profile estimation.

        Parameters
        ----------
        datasets : list of `~gammapy.datasets.MapDataset`
            Map datasets.

        Returns
        -------
        profile : `~gammapy.estimators.FluxPoints`
            Profile flux points.
        """
        datasets = Datasets(datasets=datasets)

        maps = parallel.run_multiprocessing(
            self._run_region,
            zip(repeat(datasets), self.regions),
            backend=self.parallel_backend,
            pool_kwargs=dict(processes=self.n_jobs),
            task_name="Flux profile estimation",
        )

        return FluxPoints.from_stack(
            maps=maps,
            axis=self.projected_distance_axis,
        )

    def _run_region(self, datasets, region):
        # TODO: test if it would be more efficient
        # to not compute datasets_to_fit in parallel
        # to avoid copying the full datasets
        datasets_to_fit = datasets.to_spectrum_datasets(region=region)
        for dataset_spec, dataset_map in zip(datasets_to_fit, datasets):
            dataset_spec.background.data = (
                dataset_map.npred()
                .to_region_nd_map(region, func=np.sum, weights=dataset_map.mask_safe)
                .data
            )
        datasets_to_fit.models = SkyModel(self.spectral_model, name="test-source")
        estimator = self.copy()
        estimator.n_jobs = self._n_child_jobs
        return estimator._run_flux_points(datasets_to_fit)

    def _run_flux_points(self, datasets):
        return super().run(datasets)

    @property
    def config_parameters(self):
        """Configuration parameters."""
        pars = self.__dict__.copy()
        pars = {key.strip("_"): value for key, value in pars.items()}
        pars.pop("regions")
        return pars



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import astropy.units as u
import gammapy.utils.parallel as parallel
from gammapy.estimators import FluxPointsEstimator
from gammapy.utils.testing import requires_dependency


def test_parallel_mixin():
    p = parallel.ParallelMixin()

    with pytest.raises(ValueError):
        p.parallel_backend = "wrong_name"

    with pytest.raises(ValueError):
        p.n_jobs = "5 jobs"


@requires_dependency("ray")
def test_get_multiprocessing_ray():
    assert parallel.is_ray_available()

    multiprocessing = parallel.get_multiprocessing_ray()
    assert multiprocessing.__name__ == "ray.util.multiprocessing"


def test_run_multiprocessing_wrong_method():
    def func(arg):
        return arg

    with pytest.raises(ValueError):
        parallel.run_multiprocessing(
            func, [True, True], method="wrong_name", pool_kwargs=dict(processes=2)
        )


def square(x):
    return x**2


class MyTask:
    def __init__(self):
        self.sum_squared = 0

    def __call__(self, x):
        return x**2

    def callback(self, result):
        self.sum_squared += result


def test_run_multiprocessing_simple_starmap():
    N = 10
    inputs = [(_,) for _ in range(N + 1)]

    result = parallel.run_multiprocessing(
        func=square,
        inputs=inputs,
        method="starmap",
        pool_kwargs=dict(processes=2),
    )
    assert sum(result) == N * (N + 1) * (2 * N + 1) / 6


def test_run_multiprocessing_simple_apply_async():
    N = 10
    inputs = [(_,) for _ in range(N + 1)]

    task = MyTask()

    _ = parallel.run_multiprocessing(
        func=task,
        inputs=inputs,
        method="apply_async",
        pool_kwargs=dict(processes=2),
        method_kwargs=dict(callback=task.callback),
    )
    assert task.sum_squared == N * (N + 1) * (2 * N + 1) / 6


@requires_dependency("ray")
def test_run_multiprocessing_simple_ray_starmap():
    N = 10
    inputs = [(_,) for _ in range(N + 1)]

    result = parallel.run_multiprocessing(
        func=square,
        inputs=inputs,
        method="starmap",
        pool_kwargs=dict(processes=2),
        backend="ray",
    )
    assert sum(result) == N * (N + 1) * (2 * N + 1) / 6


@requires_dependency("ray")
def test_run_multiprocessing_simple_ray_apply_async():
    N = 10
    inputs = [(_,) for _ in range(N + 1)]

    task = MyTask()

    _ = parallel.run_multiprocessing(
        func=task,
        inputs=inputs,
        method="apply_async",
        pool_kwargs=dict(processes=2),
        method_kwargs=dict(callback=task.callback),
        backend="ray",
    )
    assert task.sum_squared == N * (N + 1) * (2 * N + 1) / 6


@requires_dependency("ray")
def test_multiprocessing_manager():
    with parallel.multiprocessing_manager(
        backend="ray",
        method="apply_async",
        pool_kwargs=dict(processes=2),
        method_kwargs=dict(callback=None),
    ):
        assert parallel.BACKEND_DEFAULT == "ray"
        assert parallel.POOL_KWARGS_DEFAULT["processes"] == 2
        assert parallel.METHOD_DEFAULT == "apply_async"
        assert "callback" in parallel.METHOD_KWARGS_DEFAULT

    assert parallel.BACKEND_DEFAULT == parallel.ParallelBackendEnum.multiprocessing
    assert parallel.POOL_KWARGS_DEFAULT["processes"] == 1
    assert parallel.METHOD_DEFAULT == parallel.PoolMethodEnum.starmap
    assert "callback" not in parallel.METHOD_KWARGS_DEFAULT

    fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)
    assert fpe.parallel_backend == parallel.ParallelBackendEnum.multiprocessing

    with parallel.multiprocessing_manager(backend="ray", pool_kwargs=dict(processes=2)):
        assert fpe.parallel_backend == "ray"
        assert fpe.n_jobs == 2

        fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)
        assert fpe.parallel_backend == "ray"
    assert fpe.parallel_backend == parallel.ParallelBackendEnum.multiprocessing

    fpe = FluxPointsEstimator(
        energy_edges=[1, 3, 10] * u.TeV, n_jobs=2, parallel_backend="multiprocessing"
    )
    with parallel.multiprocessing_manager(backend="ray", pool_kwargs=dict(processes=3)):
        assert fpe.parallel_backend == "multiprocessing"
        assert fpe.n_jobs == 2



r"""
.. _expdecay-temporal-model:

ExpDecay temporal model
=======================

This model parametrises an ExpDecay time model.

.. math:: F(t) = \exp \left( \frac{t - t_{\rm{ref}}}{t0} \right)

"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
from astropy.time import Time
import matplotlib.pyplot as plt
from gammapy.modeling.models import (
    ExpDecayTemporalModel,
    Models,
    PowerLawSpectralModel,
    SkyModel,
)

t0 = "5 h"
t_ref = Time("2020-10-01")
time_range = [t_ref, t_ref + 1 * u.d]
expdecay_model = ExpDecayTemporalModel(t_ref=t_ref.mjd * u.d, t0=t0)
expdecay_model.plot(time_range)
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(
    spectral_model=PowerLawSpectralModel(),
    temporal_model=expdecay_model,
    name="expdecay_model",
)
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import astropy.units as u
from astropy.table import Table
from gammapy.maps import MapAxis, RegionGeom
from gammapy.modeling.models import Model, Models, SkyModel
from gammapy.utils.gauss import Gauss2DPDF
from gammapy.utils.scripts import make_path
from .core import SourceCatalog, SourceCatalogObject

__all__ = [
    "SourceCatalogObject1LHAASO",
    "SourceCatalog1LHAASO",
]


class SourceCatalogObject1LHAASO(SourceCatalogObject):
    """One source from the 1LHAASO catalog.

    Catalog is represented by `~gammapy.catalog.SourceCatalog1LHAASO`.
    """

    _source_name_key = "Source_Name"

    def _parse(self, name, which):
        if which in self.data["Model_a"]:
            tag = ""
        elif which in self.data["Model_b"]:
            tag = "_b"
        else:
            raise ValueError("Invalid model component name")
        is_ul = False
        value = u.Quantity(self.data[f"{name}{tag}"])
        if (
            np.isnan(value) or value == 0 * value.unit
        ) and f"{name}_ul{tag}" in self.data:
            value = self.data[f"{name}_ul{tag}"]
            is_ul = True
        return value, is_ul

    def _get(self, name, which):
        value, _ = self._parse(name, which)
        return value

    def spectral_model(self, which):
        """Spectral model as a `~gammapy.modeling.models.PowerLawSpectralModel` object.

        * ``which="KM2A"`` - Sky model for KM2A analysis only.
        * ``which="WCDA"`` - Sky model for WCDA analysis only.

        """
        pars = {
            "reference": self._get("E0", which),
            "amplitude": self._get("N0", which),
            "index": self._get("gamma", which),
        }

        errs = {
            "amplitude": self._get("N0_err", which),
            "index": self._get("gamma_err", which),
        }

        model = Model.create("PowerLawSpectralModel", "spectral", **pars)

        for name, value in errs.items():
            model.parameters[name].error = value

        return model

    def spatial_model(self, which):
        """Spatial model as a `~gammapy.modeling.models.SpatialModel` object.

        * ``which="KM2A"`` - Sky model for KM2A analysis only.
        * ``which="WCDA"`` - Sky model for WCDA analysis only.

        """
        lat_0 = self._get("DECJ2000", which)
        pars = {"lon_0": self._get("RAJ2000", which), "lat_0": lat_0, "frame": "fk5"}

        pos_err = self._get("pos_err", which)
        scale_r95 = Gauss2DPDF().containment_radius(0.95)

        errs = {
            "lat_0": pos_err / scale_r95,
            "lon_0": pos_err / scale_r95 / np.cos(lat_0),
        }

        r39, is_ul = self._parse("r39", which)
        if not is_ul:
            pars["sigma"] = r39
            model = Model.create("GaussianSpatialModel", "spatial", **pars)
        else:
            model = Model.create("PointSpatialModel", "spatial", **pars)

        for name, value in errs.items():
            model.parameters[name].error = value

        return model

    @staticmethod
    def _get_components_geom(models):
        energy_axis = MapAxis.from_energy_bounds(
            "0.1 TeV", "2000 TeV", nbin=10, per_decade=True, name="energy"
        )
        regions = [m.spatial_model.evaluation_region for m in models]
        geom = RegionGeom.from_regions(
            regions, binsz_wcs="0.05 deg", axes=[energy_axis]
        )
        return geom.to_wcs_geom()

    def sky_model(self, which="both"):
        """Sky model as a `~gammapy.modeling.models.SkyModel` object.

        * ``which="both"`` -  Create a composite template if both models are available, or, use the available one
           if only one is present.
        * ``which="KM2A"`` - Sky model for KM2A analysis if available.
        * ``which="WCDA"`` - Sky model for WCDA analysis if available.

        """
        if which == "both":
            wcda = self.sky_model(which="WCDA")
            km2a = self.sky_model(which="KM2A")
            models = [m for m in [wcda, km2a] if m is not None]
            if len(models) == 2:
                geom = self._get_components_geom(models)
                mask = geom.energy_mask(energy_max=25 * u.TeV)
                geom = geom.as_energy_true
                wcda_map = Models(wcda).to_template_sky_model(geom).spatial_model.map
                model = Models(km2a).to_template_sky_model(geom, name=km2a.name)
                model.spatial_model.map.data[mask] = wcda_map.data[mask]
                model.spatial_model.filename = f"{model.name}.fits"
                return model
            else:
                return models[0]
        else:
            _, is_ul = self._parse("N0", which)
            if is_ul:
                return None
            else:
                return SkyModel(
                    spatial_model=self.spatial_model(which),
                    spectral_model=self.spectral_model(which),
                    name=self.name,
                )


class SourceCatalog1LHAASO(SourceCatalog):
    """First LHAASO catalog.

    One source is represented by `~gammapy.catalog.SourceCatalogObject1LHAASO`.

    The data is from table 1 in the paper [1]_.

    The catalog table contains 90 rows / sources.

    References
    ----------
    .. [1] `Cao et al. (2023), "1LHAASO: The First LHAASO Catalog of Gamma-Ray Sources"
       <https://ui.adsabs.harvard.edu/abs/2023arXiv230517030C/abstract>`_
    """

    tag = "1LHAASO"
    """Catalog name"""

    description = "1LHAASO catalog from the LHAASO observatory"
    """Catalog description"""

    source_object_class = SourceCatalogObject1LHAASO

    def __init__(self, filename="$GAMMAPY_DATA/catalogs/1LHAASO_catalog.fits"):
        table = Table.read(make_path(filename))

        source_name_key = "Source_Name"

        super().__init__(table=table, source_name_key=source_name_key)

    def to_models(self, which="both"):
        """Create Models object from catalog.

        * ``which="both"`` - Use first model or create a composite template if both models are available.
        * ``which="KM2A"`` - Sky model for KM2A analysis if available.
        * ``which="WCDA"`` - Sky model for WCDA analysis if available.
        """
        models = Models()
        for _ in self:
            model = _.sky_model(which)
            if model:
                models.append(model)
        return models



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from regions import CircleSkyRegion, PointSkyRegion
from gammapy.data import DataStore, EventList
from gammapy.datasets import MapDataset, SpectrumDataset
from gammapy.makers import MapDatasetMaker, PhaseBackgroundMaker, SpectrumDatasetMaker
from gammapy.maps import MapAxis, RegionGeom, WcsGeom
from gammapy.utils.testing import requires_data

POS_CRAB = SkyCoord(83.6331, +22.0145, frame="icrs", unit="deg")
POS_VELA = SkyCoord("08h35m20.65525s", "-45d10m35.1545s", frame="icrs")


@pytest.fixture(scope="session")
def observations_cta():
    datastore_cta = DataStore.from_dir("$GAMMAPY_DATA/cta-1dc/index/gps")
    return datastore_cta.get_observations([111630])


@pytest.fixture(scope="session")
def observations_magic():
    datastore_magic = DataStore.from_dir("$GAMMAPY_DATA/magic/rad_max/data")
    return datastore_magic.get_observations(required_irf="point-like")[0]


@pytest.fixture(scope="session")
def observations_hess():
    datastore_hess = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    return datastore_hess.obs(23523)


@pytest.fixture(scope="session")
def phase_bkg_maker():
    """Example background estimator for testing."""
    return PhaseBackgroundMaker(on_phase=(0.5, 0.6), off_phase=(0.7, 1))


@requires_data()
def test_basic(phase_bkg_maker):
    assert "PhaseBackgroundMaker" in str(phase_bkg_maker)


@requires_data()
def test_run_spectrum(observations_cta, phase_bkg_maker):
    maker = SpectrumDatasetMaker()

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")
    e_true = MapAxis.from_edges(np.logspace(-0.5, 2, 11) * u.TeV, name="energy_true")

    radius = Angle(0.2, "deg")
    region = CircleSkyRegion(POS_VELA, radius)

    geom = RegionGeom.create(region=region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    obs = observations_cta[0]
    dataset = maker.run(dataset_empty, obs)
    dataset_on_off = phase_bkg_maker.run(dataset, obs)

    assert_allclose(dataset_on_off.acceptance, 0.1)
    assert_allclose(dataset_on_off.acceptance_off, 0.3)

    assert_allclose(dataset_on_off.counts.data.sum(), 28)
    assert_allclose(dataset_on_off.counts_off.data.sum(), 57)


@requires_data()
def test_run_map(observations_cta, phase_bkg_maker):
    maker = MapDatasetMaker()

    e_reco = MapAxis.from_edges(np.logspace(0, 2, 5) * u.TeV, name="energy")
    e_true = MapAxis.from_edges(np.logspace(-0.5, 2, 11) * u.TeV, name="energy_true")

    binsz = Angle(0.02, "deg")
    geom = WcsGeom.create(binsz=binsz, skydir=POS_VELA, width="2 deg", axes=[e_reco])
    dataset_empty = MapDataset.create(geom=geom, energy_axis_true=e_true)

    obs = observations_cta[0]
    dataset = maker.run(dataset_empty, obs)
    dataset_on_off = phase_bkg_maker.run(dataset, obs)

    assert_allclose(dataset_on_off.acceptance, 0.1)
    assert_allclose(dataset_on_off.acceptance_off, 0.3)

    assert_allclose(dataset_on_off.counts.data.sum(), 78)
    assert_allclose(dataset_on_off.counts_off.data.sum(), 263)


@pytest.mark.parametrize(
    "pars",
    [
        {"p_in": [[0.2, 0.3]], "p_out": [[0.2, 0.3]]},
        {"p_in": [[0.9, 0.1]], "p_out": [[0.9, 1], [0, 0.1]]},
        {"p_in": [[-0.2, 0.1]], "p_out": [[0.8, 1], [0, 0.1]]},
        {"p_in": [[0.8, 1.2]], "p_out": [[0.8, 1], [0, 0.2]]},
        {"p_in": [[0.8, 1]], "p_out": [[0.8, 1]]},
        {"p_in": [[0.2, 0.4], [0.8, 0.9]], "p_out": [[0.2, 0.4], [0.8, 0.9]]},
    ],
)
def test_check_phase_intervals(pars):
    assert_allclose(
        PhaseBackgroundMaker._check_intervals(pars["p_in"]), pars["p_out"], rtol=1e-5
    )


def test_copy_interval():
    on = (0.8, 1.2)
    off = [(0.1, 0.3), (0.4, 0.5)]
    PhaseBackgroundMaker(on_phase=on, off_phase=off)
    assert on == (0.8, 1.2)
    assert off == [(0.1, 0.3), (0.4, 0.5)]


@requires_data()
@pytest.mark.parametrize(
    "pars",
    [
        {
            "p_in": ["observations_magic", PointSkyRegion(POS_CRAB)],
            "p_out": [
                np.array([58, 19, 2, 2, 0, 0]),
                np.array([163, 46, 15, 1, 0, 0]),
            ],
        },
        {
            "p_in": ["observations_magic", CircleSkyRegion(POS_CRAB, 0.1 * u.deg)],
            "p_out": [
                np.array([23, 18, 2, 2, 0, 0]),
                np.array([51, 35, 15, 1, 0, 0]),
            ],
        },
        {
            "p_in": ["observations_hess", CircleSkyRegion(POS_CRAB, 0.1 * u.deg)],
            "p_out": [
                np.array([0, 1, 7, 4, 0, 0]),
                np.array([0, 9, 32, 12, 1, 0]),
            ],
        },
    ],
)
def test_make_counts(phase_bkg_maker, pars, request):
    maker = SpectrumDatasetMaker(
        containment_correction=False, selection=["counts", "exposure", "edisp"]
    )

    e_reco = MapAxis.from_energy_bounds(0.05, 100, nbin=6, unit="TeV", name="energy")
    e_true = MapAxis.from_energy_bounds(
        0.01, 300, nbin=15, unit="TeV", name="energy_true"
    )

    obs = request.getfixturevalue(pars["p_in"][0])
    table = obs.events.table
    table["PHASE"] = np.linspace(0, 1, len(table["TIME"]))
    obs._events = EventList(table)

    on_region = pars["p_in"][1]

    geom = RegionGeom.create(region=on_region, axes=[e_reco])
    dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=e_true)

    dataset = maker.run(dataset_empty, obs)
    dataset_on_off = phase_bkg_maker.run(dataset, obs)

    assert_allclose(
        [
            np.squeeze(dataset_on_off.counts.data),
            np.squeeze(dataset_on_off.counts_off.data),
        ],
        pars["p_out"],
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from copy import deepcopy
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from astropy.coordinates import Angle
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from gammapy.irf import EnergyDependentMultiGaussPSF, PSFKing
from gammapy.utils.testing import mpl_plot_check, requires_data


@requires_data()
class TestEnergyDependentMultiGaussPSF:
    @pytest.fixture(scope="session")
    def psf(self):
        filename = "$GAMMAPY_DATA/tests/unbundled/irfs/psf.fits"
        return EnergyDependentMultiGaussPSF.read(filename, hdu="POINT SPREAD FUNCTION")

    def test_info(self, psf):
        info_str = open(get_pkg_data_filename("./data/psf_info.txt")).read()

        print(psf.info())
        assert psf.info() == info_str

    def test_write(self, tmp_path, psf):
        psf.write(tmp_path / "tmp.fits")

        with fits.open(tmp_path / "tmp.fits", memmap=False) as hdu_list:
            assert len(hdu_list) == 2

    def test_to_table_psf(self, psf):
        energy = 1 * u.TeV
        theta = 0 * u.deg

        containment = [0.68, 0.8, 0.9]
        desired = psf.containment_radius(
            energy_true=energy, offset=theta, fraction=containment
        )

        assert_allclose(desired, [0.14775, 0.18675, 0.25075] * u.deg, rtol=1e-3)

    def test_to_unit(self, psf):
        with pytest.raises(NotImplementedError):
            psf.to_unit("deg-2")

    def test_to_psf3d(self, psf):
        rads = np.linspace(0.0, 1.0, 101) * u.deg
        psf_3d = psf.to_psf3d(rads)

        rad_axis = psf_3d.axes["rad"]
        assert rad_axis.nbin == 100
        assert rad_axis.unit == "deg"

        theta = 0.5 * u.deg
        energy = 0.5 * u.TeV

        containment = [0.68, 0.8, 0.9]
        desired = psf.containment_radius(
            energy_true=energy, offset=theta, fraction=containment
        )
        actual = psf_3d.containment_radius(
            energy_true=energy, offset=theta, fraction=containment
        )
        assert_allclose(np.squeeze(desired), actual, atol=0.005)

        # test default case
        psf_3d_def = psf.to_psf3d()
        assert psf_3d_def.axes["rad"].nbin == 66

    def test_eq(self, psf):
        psf1 = deepcopy(psf)
        assert psf1 == psf

        psf1.data[0][0] = 10
        assert not psf1 == psf

    def test_peek(self, psf):
        with mpl_plot_check():
            psf.peek()


@requires_data()
def test_psf_cta_1dc():
    filename = (
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    psf_irf = EnergyDependentMultiGaussPSF.read(filename, hdu="POINT SPREAD FUNCTION")

    # Check that PSF is filled with 0 for energy / offset where no PSF info is given.
    # This is needed so that stacked PSF computation doesn't error out,
    # trying to interpolate for observations / energies where this occurs.
    value = psf_irf.evaluate(
        energy_true=0.05 * u.TeV, rad=0.03 * u.deg, offset=4.5 * u.deg
    )
    assert_allclose(value, 0 * u.Unit("deg-2"))

    # Check that evaluation works for an energy / offset where an energy is available
    radius = psf_irf.containment_radius(
        fraction=0.68, energy_true=1 * u.TeV, offset=2 * u.deg
    )
    assert_allclose(radius, 0.052841 * u.deg, atol=1e-4)


@requires_data()
def test_get_sigmas_and_norms():
    filename = "$GAMMAPY_DATA/cta-caldb/Prod5-South-20deg-AverageAz-14MSTs37SSTs.180000s-v0.1.fits.gz"  # noqa: E501

    psf_irf = EnergyDependentMultiGaussPSF.read(filename, hdu="POINT SPREAD FUNCTION")

    value = psf_irf.evaluate(
        energy_true=1 * u.TeV, rad=0.03 * u.deg, offset=3.5 * u.deg
    )
    assert_allclose(value, 78.25826069 * u.Unit("deg-2"))


@pytest.fixture(scope="session")
def psf_king():
    return PSFKing.read("$GAMMAPY_DATA/tests/hess_psf_king_023523.fits.gz")


@requires_data()
def test_psf_king_evaluate(psf_king):
    param_off1 = psf_king.evaluate_parameters(energy_true=1 * u.TeV, offset=0 * u.deg)
    param_off2 = psf_king.evaluate_parameters(energy_true=1 * u.TeV, offset=1 * u.deg)

    assert_allclose(param_off1["gamma"].value, 1.733179, rtol=1e-5)
    assert_allclose(param_off2["gamma"].value, 1.812795, rtol=1e-5)
    assert_allclose(param_off1["sigma"], 0.040576 * u.deg, rtol=1e-5)
    assert_allclose(param_off2["sigma"], 0.040765 * u.deg, rtol=1e-5)


@requires_data()
def test_psf_king_containment_radius(psf_king):
    radius = psf_king.containment_radius(
        fraction=0.68, energy_true=1 * u.TeV, offset=0.0 * u.deg
    )

    assert_allclose(radius, 0.14575 * u.deg, rtol=1e-5)


@requires_data()
def test_psf_king_evaluate_2(psf_king):
    theta1 = Angle(0, "deg")
    theta2 = Angle(1, "deg")
    rad = Angle(1, "deg")
    # energy = Quantity(1, "TeV") match with bin number 8
    # offset equal 1 degree match with the bin 200 in the psf_table
    value_off1 = psf_king.evaluate(rad=rad, energy_true=1 * u.TeV, offset=theta1)
    value_off2 = psf_king.evaluate(rad=rad, energy_true=1 * u.TeV, offset=theta2)
    # Test that the value at 1 degree in the histogram for the energy 1 Tev and
    # theta=0 or 1 degree is equal to the one obtained from the self.evaluate_direct()
    # method at 1 degree
    assert_allclose(0.005234 * u.Unit("deg-2"), value_off1, rtol=1e-4)
    assert_allclose(0.004015 * u.Unit("deg-2"), value_off2, rtol=1e-4)


@requires_data()
def test_psf_king_write(psf_king, tmp_path):
    psf_king.write(tmp_path / "tmp.fits")
    psf_king2 = PSFKing.read(tmp_path / "tmp.fits")

    assert_allclose(
        psf_king2.axes["energy_true"].edges, psf_king.axes["energy_true"].edges
    )
    assert_allclose(psf_king2.axes["offset"].center, psf_king.axes["offset"].center)
    assert_allclose(psf_king2.data["gamma"], psf_king.data["gamma"])
    assert_allclose(psf_king2.data["sigma"], psf_king.data["sigma"])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Priors for Gammapy."""

import logging
import numpy as np
import astropy.units as u
from scipy.stats import norm, uniform, loguniform, gennorm
from gammapy.modeling import PriorParameter, PriorParameters
from .core import ModelBase

__all__ = [
    "GaussianPrior",
    "GeneralizedGaussianPrior",
    "UniformPrior",
    "LogUniformPrior",
    "Prior",
]

log = logging.getLogger(__name__)


def _build_priorparameters_from_dict(data, default_parameters):
    """Build a `~gammapy.modeling.PriorParameters` object from input dictionary and default prior parameter values."""
    par_data = []

    input_names = [_["name"] for _ in data]

    for par in default_parameters:
        par_dict = par.to_dict()
        try:
            index = input_names.index(par_dict["name"])
            par_dict.update(data[index])
        except ValueError:
            log.warning(
                f"PriorParameter '{par_dict['name']}' not defined in YAML file."
                f" Using default value: {par_dict['value']} {par_dict['unit']}"
            )
        par_data.append(par_dict)

    return PriorParameters.from_dict(par_data)


class Prior(ModelBase):
    """Prior abstract base class. For now, see existing examples of type of priors:

    - `GaussianPrior`
    - `UniformPrior`
    - `LogUniformPrior`
    """

    _unit = ""

    def __init__(self, **kwargs):
        # Copy default parameters from the class to the instance
        default_parameters = self.default_parameters.copy()

        for par in default_parameters:
            value = kwargs.get(par.name, par)
            if not isinstance(value, PriorParameter):
                par.quantity = u.Quantity(value)
            else:
                par = value

            setattr(self, par.name, par)

        _weight = kwargs.get("weight", None)

        if _weight is not None:
            self._weight = _weight
        else:
            self._weight = 1

    @property
    def parameters(self):
        """Prior parameters as a `~gammapy.modeling.PriorParameters` object."""
        return PriorParameters(
            [getattr(self, name) for name in self.default_parameters.names]
        )

    def __init_subclass__(cls, **kwargs):
        # Add priorparameters list on the model sub-class (not instances)
        cls.default_parameters = PriorParameters(
            [_ for _ in cls.__dict__.values() if isinstance(_, PriorParameter)]
        )

    def _inverse_cdf(self, value):
        """Return inverse CDF for prior."""
        return self._random_variable.ppf(value)

    @property
    def weight(self):
        """Weight multiplied to the prior when evaluated."""
        return self._weight

    @weight.setter
    def weight(self, value):
        self._weight = value

    def __call__(self, value):
        """Call evaluate method."""
        # assuming the same unit as the PriorParameter here
        kwargs = {par.name: par.value for par in self.parameters}
        return self.weight * self.evaluate(value.value, **kwargs)

    def to_dict(self, full_output=False):
        """Create dictionary for YAML serialisation."""
        tag = self.tag[0] if isinstance(self.tag, list) else self.tag
        params = self.parameters.to_dict()

        if not full_output:
            for par, par_default in zip(params, self.default_parameters):
                init = par_default.to_dict()
                for item in [
                    "min",
                    "max",
                    "error",
                ]:
                    default = init[item]

                    if par[item] == default or (
                        np.isnan(par[item]) and np.isnan(default)
                    ):
                        del par[item]

        data = {"type": tag, "parameters": params, "weight": self.weight}

        if self.type is None:
            return data
        else:
            return {self.type: data}

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Get prior parameters from dictionary."""
        kwargs = {}

        key0 = next(iter(data))
        if key0 in ["prior"]:
            data = data[key0]
        if data["type"] not in cls.tag:
            raise ValueError(
                f"Invalid model type {data['type']} for class {cls.__name__}"
            )

        priorparameters = _build_priorparameters_from_dict(
            data["parameters"], cls.default_parameters
        )
        kwargs["weight"] = data["weight"]
        return cls.from_parameters(priorparameters, **kwargs)


class GaussianPrior(Prior):
    """One-dimensional Gaussian Prior.

    Parameters
    ----------
    mu : float, optional
        Mean of the Gaussian distribution.
        Default is 0.
    sigma : float, optional
        Standard deviation of the Gaussian distribution.
        Default is 1.
    """

    tag = ["GaussianPrior"]
    _type = "prior"
    mu = PriorParameter(name="mu", value=0)
    sigma = PriorParameter(name="sigma", value=1)

    @staticmethod
    def evaluate(value, mu, sigma):
        """Evaluate the Gaussian prior."""
        rv = norm(mu, sigma)
        return -2 * rv.logpdf(value)

    @property
    def _random_variable(self):
        """Return random variable object for prior."""
        return norm(self.mu.value, self.sigma.value)


class UniformPrior(Prior):
    """Uniform Prior.

    Returns 2log(max-min) if the parameter value is in [min, max].
    Returns inf, otherwise.
    Only well defined for finite values of min and max.

    Parameters
    ----------
    min : float, optional
        Minimum value.
        Default is 0.
    max : float, optional
        Maximum value.
        Default is 1.
    """

    tag = ["UniformPrior"]
    _type = "prior"
    min = PriorParameter(name="min", value=0.0, unit="")
    max = PriorParameter(name="max", value=1.0, unit="")

    @staticmethod
    def evaluate(value, min, max):
        """Evaluate the uniform prior."""
        rv = uniform(min, max - min)
        return -2 * rv.logpdf(value)

    @property
    def _random_variable(self):
        """Return random variable object for prior."""
        return uniform(self.min.value, self.max.value - self.min.value)


class LogUniformPrior(Prior):
    """LogUniform Prior.

    Equivalent to a uniform prior on the log of the parameter

    Parameters
    ----------
    min : float, optional
        Minimum value. Default is 1e-14.
    max : float, optional
        Maximum value. Default is 1e-10.
    """

    tag = ["LogUniformPrior"]
    _type = "prior"
    min = PriorParameter(name="min", value=1e-14, unit="")
    max = PriorParameter(name="max", value=1e-10, unit="")

    @staticmethod
    def evaluate(value, min, max):
        """
        Evaluate the likelihood penalization term (hence -2*).
        Note that this is currently a different scaling that the Uniform or Gaussian priors.
        With current implementation the TS of a source with/without LogUniform prior would be different... TBD
        """
        rv = loguniform(min, max)
        return -2 * rv.logpdf(value)

    @property
    def _random_variable(self):
        """Return random variable object for prior."""
        return loguniform(self.min.value, self.max.value)


class GeneralizedGaussianPrior(Prior):
    """One-dimensional Generalized Gaussian Prior.

    Parameters
    ----------
    mu : float, optional
        Mean of the Gaussian distribution.
        Default is 0.
    sigma : float, optional
        Standard deviation of the Gaussian distribution.
        Default is 1.
    eta : `float`, optional
        eta is a shape parameter
        For eta=1 it is identical to a Laplace distribution (scaled by sqrt(2)).
        For eta=0.5 it is identical to a normal distribution.
        Default is 0.5.
    """

    tag = ["GeneralizedGaussianPrior"]
    _type = "prior"
    mu = PriorParameter(name="mu", value=0)
    sigma = PriorParameter(name="sigma", value=1)
    eta = PriorParameter(name="eta", value=0.5)

    @staticmethod
    def evaluate(value, mu, sigma, eta):
        """Evaluate the Gaussian prior."""
        rv = gennorm(beta=1.0 / eta, loc=mu, scale=sigma * np.sqrt(2))
        return -2 * rv.logpdf(value)

    @property
    def _random_variable(self):
        """Return random variable object for prior."""
        return gennorm(
            beta=1.0 / self.eta.value,
            loc=self.mu.value,
            scale=self.sigma.value * np.sqrt(2),
        )
